#!/usr/bin/env python
"""Infer a schema from a rdftab

"""
from linkml.utils.schemaloader import load_raw_schema
from typing import Union, List, Set, Optional
from linkml.generators.yamlgen import YAMLGenerator

import logging
import click
import yaml
import csv
import time

from linkml_runtime.linkml_model.meta import SchemaDefinition
from linkml.utils.generator import Generator, shared_arguments
from dateutil.parser import parse

CURIE = str

RANGE_COMPOSITION_TABLE = {
    ('xsd:integer', 'xsd:float'): 'number',
    ('owl:Thing', 'owl:Class'): 'owl:Thing',
    ('owl:ObjectProperty', 'owl:TransitiveProperty'): 'owl:ObjectProperty',
    ('owl:ObjectProperty', 'owl:SymmetrucProperty'): 'owl:ObjectProperty'
}

OMIT_PRED_IN = ['rdf', 'swrl']
OMIT_CLASS_IN = ['rdf', 'swrl', 'UO']

def remove_prefix(id: str) -> str:
    """
    get the local part of a CURIE or URI
    """
    if 'http' in id:
        if id.startswith('<'):
            id = id.replace('<', '').replace('>', '')
        if '#' in id:
            parts = id.split('#')
        elif '/'  in id:
            parts = id.split('.')
        else:
            logging.error(f'ID: {id}')
            parts = [id]
    else:
        parts = id.split(':')
        if len(parts) > 2:
            logging.warning(f'ID has > 2 parts: {id}')
    local_id = parts[-1]
    if '#' in local_id:
        parts = local_id.split('#')
        local_id = parts[-1]
    if '/' in local_id:
        parts = local_id.split('/')
        local_id = parts[-1]
    return local_id


def is_literal(x: CURIE) -> bool:
    """
    True if the URI is an xsd type
    """
    return x.startswith('xsd:')

def is_thing_or_individual(x: CURIE) -> bool:
    """
    True if the URI is owl Thing or NamedIndividual
    """
    return x == 'owl:Thing' or x == 'owl:NamedIndividual'

def condense_range_pair(r1: CURIE, r2: CURIE) -> Optional[CURIE]:
    """
    Infer a parent range from a pair of ranges

    TODO: use the ontology to do this
    """
    if r1 == r2:
        return r1
    if (r1,r2) in RANGE_COMPOSITION_TABLE:
        return RANGE_COMPOSITION_TABLE[(r1,r2)]
    elif (r2,r1) in RANGE_COMPOSITION_TABLE:
        return RANGE_COMPOSITION_TABLE[(r2,r1)]
    else:
        if is_literal(r1) != is_literal(r2):
            logging.warning(f'Trying to combine literal and non-literal: {r1} {r2}')
        if is_thing_or_individual(r1):
            return r2
        elif is_thing_or_individual(r2):
            return r1
        else:
            logging.warning(f'Logic not implemented to condense: {r1} {r2}')
            return None

def condense_ranges(cset: Set[CURIE], classes: dict) -> Optional[str]:
    """
    Condense a set of ranges into one range that subsumes them all

    TODO: use the ontology to do this
    """
    todo = list(cset)
    r = todo.pop()
    while len(todo) > 0:
        r2 = todo.pop()
        r = condense_range_pair(r, r2)
        if r is None:
            return None
    return remove_prefix(r)

def new_cls(u: CURIE) -> dict:
    """
    Create stub linkml ClassDefinition dict
    """
    return {'class_uri': u, 'slots': [], 'slot_usage': {}}

def remove_angle_brackets(id: str) -> str:
    if id.startswith('<'):
        id = id.replace('<', '').replace('>', '')
    if id.startswith('obo:http'):
        id = id.replace('obo:', '')
    return id

def set_range(slot: dict, r: str, classes: dict) -> None:
    if r in classes:
        slot['range'] = r
    else:
        if 'range' in slot:
            del slot['range']
        logging.warning(f'Omitting range: {r} as there is no class definitions')

def infer_model_from_predicate_summary(tsvfile: str, sep="\t",
                                       schema_name: str = 'example',
                                       count_threshold: int=2,
                                       include_counts = True,
                                       base_uri: str = None,
                                       schema_description: str = 'TODO',
                                       enum_columns: List[str]=[],
                                       enum_threshold=0.1,
                                       max_enum_size=50) -> dict:
    """
    Given a predicate summary table, infer a linkml schema
    """
    with open(tsvfile, newline='') as tsvfile:
        rr = csv.DictReader(tsvfile, delimiter=sep)
        slots = {}
        slot_values = {}
        n = 0
        enums = {}
        slot_usage = {}
        classes = {}
        slot_ranges = {}
        for row in rr:
            pred = remove_angle_brackets(row['predicate'])
            domain = row['subject_type']
            range = row['object_type']
            num = int(row['num_statements'])
            predname = row['predicate_label']
            is_multivalued = int(row['is_multivalued']) != 0
            if num < count_threshold:
                continue
            # TODO: make configurable
            if pred == 'IAO:0000111':
                # https://github.com/EnvironmentOntology/envo/issues/1187
                predname = 'editor_preferred_term'
            # TODO: make configurable
            if pred.startswith('oio:http'):
                pred = pred.replace('oio:http', 'http')
            dc = remove_prefix(domain)
            rc = remove_prefix(range)
            if predname is None or predname == '':
                sn = remove_prefix(pred)
            else:
                sn = predname
            if sn is None or sn == '':
                raise Exception(f'Null p for {row}')
            if sn.isnumeric():
                logging.warning(f'Ignoring: {pred}, no name {predname}')
                continue
            sn = sn.replace(' ', '_')
            #p = p.lower()
            if dc not in classes:
                classes[dc] = new_cls(domain)
            c = classes[dc]
            if rc not in classes:
                classes[rc] = new_cls(range)
            if sn not in slots:
                slots[sn] = {'slot_uri': pred, 'multivalued': is_multivalued}
                slot_ranges[sn] = set()
            generic_slot = slots[sn]
            if generic_slot['slot_uri'] != pred:
                if 'exact_mappings' not in generic_slot:
                    generic_slot['exact_mappings'] = []
                if pred not in generic_slot['exact_mappings']:
                    generic_slot['exact_mappings'].append(pred)
                logging.warning(f'Collapsing {pred} and {generic_slot["slot_uri"]} into {sn}')
            slot_ranges[sn].add(range)
            if sn not in c['slots']:
                c['slots'].append(sn)
                c['slot_usage'][sn] = {'range': set(), 'count': 0}
            c['slot_usage'][sn]['range'].add(range)
            c['slot_usage'][sn]['count'] += num
        # filter useless classes
        classes = {cn:c for cn,c in classes.items() if len(c['slots']) > 0}
        for sn, ranges in slot_ranges.items():
            r = condense_ranges(ranges, classes)
            if r is not None:
                logging.error(f'Set range for slot {sn} to {r}')
                set_range(slots[sn], r, classes)
            else:
                slots[sn]['todos'] = ['define range']
        for cn, c in classes.items():
            redundant = []
            for sn, class_slot in c['slot_usage'].items():
                r = condense_ranges(class_slot['range'], classes)
                if r is None:
                    del class_slot['range']
                    class_slot['todos'] = ['define range']
                else:
                    logging.error(f'Setting range for slot {cn}.{sn} in {class_slot} to {r}')
                    set_range(class_slot, r, classes)
                    logging.error(f'Set range for slot {cn}.{sn} in {class_slot} to {r}')
                n = class_slot['count']
                del class_slot['count']
                # check for redundancy
                is_redundant = True
                if sn in slots:
                    generic_slot = slots[sn]
                    for k, v in class_slot.items():
                        if k in generic_slot:
                            if generic_slot[k] != v:
                                logging.error(f'OVERIDE {sn} in {cn} slot property: {k} :: {v} >> {generic_slot[k]}')
                                is_redundant = False
                        else:
                            logging.error(f'NONRED {sn} in {cn} slot property: {k} not in {generic_slot}')
                            is_redundant = False
                else:
                    raise Exception(f'No slot: {sn}')
                if include_counts:
                    class_slot['comments'] = [f'Usages: {n}']
                if is_redundant:
                    redundant.append(sn)
                    logging.error(f'deleting redundant usage for {cn}.{sn}')
                else:
                    logging.error(f'keeping nonredundant usage for {cn}.{sn}')
            for sn in redundant:
                del c['slot_usage'][sn]
    # TODO: infer if required; infer multivalued
    if base_uri is None:
        base_uri = f'https://w3id.org/{schema_name}'
    schema = {
        'id': base_uri,
        'name': schema_name,
        'description': schema_description,
        'imports': ['linkml:types'],
        'prefixes': {
            'linkml': 'https://w3id.org/linkml/',
            schema_name: f'{base_uri}/'
        },
        'default_prefix': schema_name,
        'types': {},
        'classes': classes,
        'slots': slots,
        'enums': enums
    }
    return schema

@click.group()
def main():
    pass

@main.command()
@click.argument('tsvfile') ## input TSV (must have column headers
@click.option('--schema_name', '-n', default='example', help='Schema name')
@click.option('--sep', '-s', default='\t', help='separator')
@click.option('--base-uri', '-b', help='base URI')
@click.option('--schema-description', '-D', help='schema description')
@click.option('--enum-columns', '-E', multiple=True, help='column that is forced to be an enum')
@click.option('--include-counts/--no-include-counts', default=True, help='set to include counts in slot_usage')
def preds2model(tsvfile, **args):
    """
    Infer a model from a predicate summary TSV

    Pre-processing:

    sqlite3 db/foo.db -cmd '.headers on' -cmd '.separator "\t"' "select * from predicate_summary_labeled"

    """
    s = infer_model_from_predicate_summary(tsvfile, **args)
    ys = yaml.dump(s, default_flow_style=False, sort_keys=False)
    schema = load_raw_schema(ys)
    G = YAMLGenerator(schema)
    print(G.serialize())

if __name__ == '__main__':
    main()



import click
import pandas as pd
from linkml_runtime.utils.schemaview import SchemaView

# specify normalized_dist and dist type?
blank_row = {"text": "", "title": "", "meaning": "", "match_id": "", "match_pref_lab": "", "match_type": "",
             "match_val": "", "cosine": "", "curated_id": "", "curated_pref_lab": "", "curated_type": "",
             "curated_val": "", "curation_notes": "", }
br_keys = list(blank_row.keys())
br_keys.sort()


@click.command()
@click.option('--modelfile', type=click.Path(exists=True), required=True)
@click.option('--tsv_out', type=click.Path(), default="enums_to_curateable.tsv", show_default=True)
@click.option('--enum', required=True)
def enums_to_curateable(modelfile, enum, tsv_out):
    row_list = []
    mschema = SchemaView(modelfile)
    menum = mschema.get_enum(enum)
    mpvs = menum.permissible_values
    for apvt, apvd in mpvs.items():
        current_blank = blank_row.copy()
        apvd_annotations = apvd.annotations
        apvd_anno_keys = list(apvd_annotations.keys())
        for i in br_keys:
            if i in apvd_anno_keys:
                current_blank[i] = apvd_annotations[i].value
            elif i in apvd:
                current_blank[i] = apvd[i]
        row_list.append(current_blank)
        row_frame = pd.DataFrame(row_list)
    print(row_frame)
    row_frame.to_csv(tsv_out, sep="\t", index=False)


if __name__ == '__main__':
    enums_to_curateable()



import logging

import click
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SlotDefinitionName
from linkml_runtime.utils.introspection import package_schemaview
from linkml_runtime.utils.schema_as_dict import schema_as_dict, schema_as_yaml_dump

from schema_automator.utils.instance_extractor import InstanceView
from schema_automator.utils.schemautils import minify_schema


@click.command()
@click.argument('elements', nargs=-1)
@click.option('--mergeimports/--no-mergeimports', help="Merge imports closure")
@click.option('--input', '-i', help="Data input")
@click.option('--output', '-o', help="Path to saved yaml schema")
@click.option("-v", "--verbose", count=True)
def cli(elements, input, mergeimports, output, verbose: int, **args):
    """
    Extract from instance data
    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    metamodel_sv = package_schemaview('linkml_runtime.linkml_model.meta')
    input_sv = SchemaView(input)
    #schema = yaml_loader.load(source=input, target_class=SchemaDefinition)
    #print(yaml_dumper.dumps(root))
    if mergeimports:
        input_sv.merge_imports()
    input_schema = input_sv.schema
    iv = InstanceView(root=input_schema, schemaview=metamodel_sv)
    logging.info(f'Creating index')
    iv.create_index()
    subschema = iv.extract(list(elements),
                           preserve_slots=[SlotDefinitionName('name'),
                                           SlotDefinitionName('prefixes'),
                                           ])
    ys = schema_as_yaml_dump(subschema)
    if output:
        with open(output, 'w') as stream:
            stream.write(ys)
    else:
        print(ys)


if __name__ == '__main__':
    cli()






import copy
import logging
from typing import Union, Optional
from deprecated.classic import deprecated


import yaml
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.schema_as_dict import schema_as_dict


def write_schema(schema: Union[dict, SchemaDefinition], output: Optional[str]):
    """
    Convenience method for writing a schema to stdout or to a file

    :param schema:
    :param output: if None, then write to stdout
    :return:
    """
    if isinstance(schema, SchemaDefinition):
        sdict = schema_as_dict(schema)
    else:
        sdict = schema
    ys = yaml.safe_dump(sdict, sort_keys=False)
    if output:
        with open(output, 'w', encoding='UTF-8') as stream:
            stream.write(ys)
    else:
        print(ys)


@deprecated("Replaced by linkml.runtime.utils.schema_as_dict")
def minify_schema(obj: Union[dict, SchemaDefinition]) -> dict:
    # TODO prefixes
    if isinstance(obj, SchemaDefinition):
        yd = yaml_dumper.dumps(obj)
        obj = yaml.safe_load(yd)
    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(v, dict) and 'name' in v and v['name'] == k:
                del v['name']
            minify_schema(v)
    elif isinstance(obj, list):
        for v in obj:
            minify_schema(v)
    else:
        None
    return obj


@deprecated("Replaced by schemaview")
def merge_schemas(schemas, nomerge_enums_for=[]):
    schema = copy.deepcopy(schemas[0])
    for s in schemas:
        for n,x in s['classes'].items():
            if n not in schema['classes']:
                schema['classes'][n] = x
        for n,x in s['slots'].items():
            if n not in schema['slots']:
                schema['slots'][n] = x
            else:
                cur_slot = schema['slots'][n]
                if 'range' in cur_slot and 'range' in x:
                    if cur_slot['range'] != x['range']:
                        logging.error(f'Inconsistent ranges: {cur_slot} vs {x}')
                        if cur_slot['range'] == 'string':
                            cur_slot['range'] = x['range']
                elif 'range' not in cur_slot:
                    cur_slot['range'] = x['range']
        for n,x in s['types'].items():
            if n not in schema['types']:
                schema['types'][n] = x
            else:
                None # TODO
        for n,x in s['enums'].items():
            if n not in schema['enums']:
                schema['enums'][n] = x
            else:
                None # TODO
    return schema


"""
Command Line Interface to Schema Automator
-----------------------------

"""
import logging
import os
import click

import yaml
from linkml_runtime.linkml_model import SchemaDefinition
from oaklib.selector import get_resource_from_shorthand, get_implementation_from_shorthand

from schema_automator import JsonLdAnnotator
from schema_automator.annotators.schema_annotator import SchemaAnnotator
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from schema_automator.importers.dosdp_import_engine import DOSDPImportEngine
from schema_automator.generalizers.json_instance_generalizer import JsonDataGeneralizer
from schema_automator.importers.jsonschema_import_engine import JsonSchemaImportEngine
from schema_automator.importers.owl_import_engine import OwlImportEngine
from schema_automator.generalizers.rdf_data_generalizer import RdfDataGeneralizer
from schema_automator.utils.schemautils import minify_schema, write_schema

input_option = click.option(
    "-i",
    "--input",
    help="path to input file"
)
output_option = click.option(
    "-o",
    "--output",
    help="path to output file."
)
schema_name_option = click.option(
    '--schema-name',
    '-n',
    default='example',
    show_default=True,
    help='Schema name')


@click.group()
@click.option("-v", "--verbose",
              count=True,
              help="Set the level of verbosity")
@click.option("-q", "--quiet",
              help="Silence all diagnostics")
def main(verbose: int, quiet: bool):
    """Run the LinkML Schema Automator Command Line.

    A subcommand must be passed, for example:

        schemauto -v SUBCOMMAND [OPTIONS] ARGUMENTS
    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if quiet:
        logging.basicConfig(level=logging.ERROR)


@main.command()
@click.argument('tsvfile')  # input TSV (must have column headers
@output_option
@schema_name_option
@click.option('--class-name', '-c', default='example', help='Core class name in schema')
@click.option('--column-separator', '-s', default='\t', help='separator')
@click.option('--downcase-header/--no-downcase-header', default=False, help='if true make headers lowercase')
@click.option('--enum-columns', '-E', multiple=True, help='column that is forced to be an enum')
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
def generalize_tsv(tsvfile, output, class_name, schema_name, **kwargs):
    """
    Generalizes from a single TSV file to a single-class schema

    See :ref:`generalizers` for more on the generalization framework

    Example:

        schemauto generalize-tsv --class-name Person --schema-name PersonInfo my/data/persons.tsv
    """
    ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert(tsvfile, class_name=class_name, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('tsvfiles', nargs=-1)  # input TSV (must have column headers
@output_option
@schema_name_option
@click.option('--column-separator', '-s', default='\t', help='separator')
@click.option('--downcase-header/--no-downcase-header', default=False, help='if true make headers lowercase')
@click.option('--infer-foreign-keys/--no-infer-foreign-keys', default=False, help='infer ranges/foreign keys')
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
def generalize_tsvs(tsvfiles, output, schema_name, **kwargs):
    """
    Generalizes from a multiple TSV files to a multi-class schema

    See :ref:`generalizers` for more on the generalization framework

    This uses :ref:`CsvDataGeneralizer.convert_multiple`

    Example:

        schemauto generalize-tsvs --class-name Person --schema-name PersonInfo my/data/*.tsv
    """
    ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert_multiple(tsvfiles, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('dpfiles', nargs=-1) ## input DOSDPs
@output_option
@schema_name_option
@click.option('--range-as-enum/--no-range-as-enums',
              default=True,
              help="Model range ontology classes as enums")
def import_dosdps(dpfiles, output, **args):
    """
    Imports DOSDP pattern YAML to a LinkML schema

    See :ref:`importers` for more on the importers framework
    """
    ie = DOSDPImportEngine()
    schema = ie.convert(dpfiles, **args)
    write_schema(schema, output)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@click.option('--container-class-name', default='Container', help="name of root class")
@click.option('--format', '-f', default='json', help="json or yaml (or json.gz or yaml.gz) or frontmatter")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def generalize_json(input, output, schema_name, format, omit_null, **kwargs):
    """
    Generalizes from a JSON file to a schema

    See :ref:`generalizers` for more on the generalization framework

    Example:

        schemauto generalize-json my/data/persons.json
    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    schema = ie.convert(input, format=format, **kwargs)
    write_schema(schema, output)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@click.option('--format', '-f', default='json', help='JSON Schema format - yaml or json')
def import_json_schema(input, output, schema_name, format, **args):
    """
    Imports from JSON Schema to LinkML

    See :ref:`importers` for more on the importer framework

    Example:

        schemauto import-json-schema my/schema/personinfo.schema.json
    """
    loader = JsonSchemaImportEngine()
    schema = loader.load(input, name=schema_name, format=format)
    write_schema(schema, output)


@main.command()
@click.argument('owlfile')
@output_option
@schema_name_option
@click.option('--identifier', '-I', help="Slot to use as identifier")
@click.option('--model-uri', help="Model URI prefix")
@click.option('--output', '-o', help="Path to saved yaml schema")
def import_owl(owlfile, output, **args):
    """
    Import an OWL ontology to LinkML

    Note this works best for "schema-style" ontologies

    See :ref:`importers` for more on the importer framework

    Note: input must be in functional syntax
    """
    sie = OwlImportEngine()
    schema = sie.convert(owlfile, **args)
    write_schema(schema, output)


@main.command()
@click.argument('rdffile')
@output_option
@click.option('--dir', '-d', required=True)
def generalize_rdf(rdffile, dir, output, **args):
    """
    Generalizes from an RDF file to a schema

    See :ref:`generalizers` for more on the generalization framework

    The input must be in turtle

    Example:

        schemauto generalize-json my/data/persons.ttl
    """
    sie = RdfDataGeneralizer()
    if not os.path.exists(dir):
        os.makedirs(dir)
    schema_dict = sie.convert(rdffile, dir=dir, format='ttl')
    write_schema(schema_dict, output)


@main.command()
@click.argument('schema')
@click.option('--curie-only/--no-curie-only',
              default=False,
              show_default=True,
              help="if set, only use results that are mapped to CURIEs")
@click.option('--input', '-i', help="OAK input ontology selector")
@output_option
def annotate_schema(schema: str, input: str, output: str, curie_only: bool, **args):
    """
    Annotate all elements of a schema

    Requires Bioportal API key
    """
    impl = get_implementation_from_shorthand(input)
    logging.basicConfig(level=logging.INFO)
    annr = SchemaAnnotator(impl)
    schema = annr.annotate_schema(schema, curie_only=curie_only)
    write_schema(schema, output)


@main.command()
@click.argument('schema')
@output_option
def annotate_using_jsonld(schema: str, output: str, **args):
    """
    Annotates a schema using a Json-LD context file
    """
    logging.basicConfig(level=logging.INFO)
    annr = JsonLdAnnotator()
    schemadef = SchemaDefinition(schema)
    annr.annotate_schema(schemadef)
    write_schema(schemadef, output)


if __name__ == "__main__":
    main()



from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from schema_automator.generalizers.json_instance_generalizer import JsonDataGeneralizer
from schema_automator.generalizers.rdf_data_generalizer import RdfDataGeneralizer


import click
import logging
import yaml
from typing import Union, Dict, Tuple, List, Any
from collections import defaultdict

from linkml_runtime.linkml_model import SchemaDefinition
from funowl.converters.functional_converter import to_python
from funowl import *

from dataclasses import dataclass
from schema_automator.importers.import_engine import ImportEngine
from schema_automator.utils.schemautils import write_schema


@dataclass
class OwlImportEngine(ImportEngine):
    """
    An ImportEngine that takes schema-style OWL and converts it to a LinkML schema
    """
    mappings: dict = None
    include_unmapped_annotations = False

    def convert(self, file: str, name: str = None, model_uri: str = None, identifier: str = None, **kwargs) -> \
            SchemaDefinition:
        """
        Converts an OWL schema-style ontology

        :param file:
        :param name:
        :param model_uri:
        :param identifier:
        :param kwargs:
        :return:
        """
        self.mappings = {}
        doc = to_python(file)
        ontology = doc.ontology
        if len(ontology.axioms) == 0:
            raise Exception(f'Empty ontology in {file} (note: ontologies must be in functional syntax)')
        prefixes = doc.prefixDeclarations
        self.prefixes = prefixes
        if model_uri is None:
            model_uri = f'https://w3id.org/{name}/'
        if name is None:
            name = self.iri_to_name(ontology.iri)
        classes = {}
        slots = {}
        enums = {}
        types = {}
        schema_dict = {
            'id': f'{ontology.iri}',
            'name': name,
            'description': name,
            'imports': ['linkml:types'],
            'prefixes': {
                'linkml': 'https://w3id.org/linkml/',
                name: model_uri,
            },
            'default_prefix': name,
            'types': types,
            'classes': classes,
            'slots': slots,
            'enums': enums
        }
        self.schema = schema_dict
        isamap = defaultdict(set)
        slot_isamap = defaultdict(set)
        slot_usage_map = defaultdict(dict)
        single_valued_slots = set()
        for a in ontology.axioms:
            logging.debug(f'Axiom: {a}')
            if isinstance(a, SubClassOf):
                if isinstance(a.subClassExpression, Class):
                    def set_slot_usage(p, k, v):
                        if p not in slot_usage_map[child]:
                            slot_usage_map[child][p] = {}
                        slot_usage_map[child][p][k] = v
                    def set_cardinality(p, min_card, max_card):
                        if max_card is not None:
                            if max_card == 1:
                                set_slot_usage(p, 'multivalued', False)
                            elif max_card > 1:
                                set_slot_usage(p, 'multivalued', True)
                        if min_card is not None:
                            if min_card == 1:
                                set_slot_usage(p, 'required', True)
                            elif min_card == 0:
                                set_slot_usage(p, 'required', False)
                            else:
                                set_slot_usage(p, 'multivalued', True)
                    child = self.iri_to_name(a.subClassExpression)
                    if isinstance(a.superClassExpression, Class):
                        parent = self.iri_to_name(a.superClassExpression)
                        isamap[child].add(parent)
                    elif isinstance(a.superClassExpression, DataExactCardinality):
                        x = a.superClassExpression
                        card = x.card
                        p = self.iri_to_name(x.dataPropertyExpression)
                        set_cardinality(p, card, card)
                    elif isinstance(a.superClassExpression, ObjectExactCardinality):
                        x = a.superClassExpression
                        card = x.card
                        p = self.iri_to_name(x.objectPropertyExpression)
                        set_cardinality(p, card, card)
                    elif isinstance(a.superClassExpression, ObjectMinCardinality):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.objectPropertyExpression)
                        set_cardinality(p, x.min_, None)
                    elif isinstance(a.superClassExpression, DataMinCardinality):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.dataPropertyExpression)
                        set_cardinality(p, x.min_, None)
                    elif isinstance(a.superClassExpression, ObjectMaxCardinality):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.objectPropertyExpression)
                        set_cardinality(p, None, x.max_)
                    elif isinstance(a.superClassExpression, DataMaxCardinality):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.dataPropertyExpression)
                        set_cardinality(p, None, x.max_)
                    elif isinstance(a.superClassExpression, ObjectAllValuesFrom):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.objectPropertyExpression)
                        if isinstance(x.classExpression, Class):
                            set_slot_usage(p, 'range', self.iri_to_name(x.classExpression))
                        else:
                            logging.error(f'Cannot yet handle anonymous ranges: {x.classExpression}')
                    elif isinstance(a.superClassExpression, ObjectSomeValuesFrom):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.objectPropertyExpression)
                        set_cardinality(p, 1, None)
                    elif isinstance(a.superClassExpression, DataSomeValuesFrom):
                        x = a.superClassExpression
                        if len(x.dataPropertyExpressions) == 1:
                            p = self.iri_to_name(x.dataPropertyExpressions[0])
                            set_cardinality(p, 1, None)
                        else:
                            logging.error(f'Cannot handle multiple data property expressions: {x}')
                    elif isinstance(a.superClassExpression, DataAllValuesFrom):
                        x = a.superClassExpression
                        if len(x.dataPropertyExpressions) == 1:
                            p = self.iri_to_name(x.dataPropertyExpressions[0])
                            r = x.dataRange
                            if isinstance(r, DataOneOf):
                                logging.error(f'TODO: enum for {r}')
                            elif isinstance(r, Datatype):
                                set_slot_usage(p, 'range', r)
                            else:
                                logging.error(f'Cannot handle range of {r}')
                        else:
                            logging.error(f'Cannot handle multiple data property expressions: {x}')
                    elif isinstance(a.superClassExpression, DataHasValue):
                        x = a.superClassExpression
                        p = self.iri_to_name(x.dataPropertyExpression)
                        #if p not in slot_usage_map[child]:
                        #    slot_usage_map[child][p] = {}
                        lit = x.literal.v
                        if isinstance(lit, TypedLiteral):
                            lit = lit.literal
                        set_slot_usage(p, 'equals_string', str(lit))
                        #slot_usage_map[child][p]['equals_string'] = str(lit)
                    else:
                        logging.error(f"cannot handle anon parent classes for {a}")
                else:
                    logging.error(f"cannot handle anon child classes for {a}")
            # https://github.com/hsolbrig/funowl/issues/19
            if isinstance(a, SubObjectPropertyOf):
                sub = a.subObjectPropertyExpression.v
                if isinstance(sub, ObjectPropertyExpression) and isinstance(sub.v, ObjectProperty):
                    child = self.iri_to_name(sub.v)
                    sup = a.superObjectPropertyExpression.v
                    if isinstance(sup, ObjectPropertyExpression) and isinstance(sup.v, ObjectProperty):
                        parent = self.iri_to_name(sup.v)
                        slot_isamap[child].add(parent)
                    else:
                        logging.error(f"cannot handle anon object parent properties for {a}")
                else:
                    logging.error(f"cannot handle anon object child properties for {a}")
            if isinstance(a, SubDataPropertyOf):
                sub = a.subDataPropertyExpression.v
                if isinstance(sub, DataProperty):
                    child = self.iri_to_name(sub)
                    sup = a.superDataPropertyExpression.v
                    if isinstance(sup, DataProperty):
                        parent = self.iri_to_name(sup)
                        slot_isamap[child].add(parent)
                    else:
                        logging.error(f"cannot handle anon data parent properties for {a}")
                else:
                    logging(f"cannot handle anon data child properties for {a}")
            if isinstance(a, SubAnnotationPropertyOf):
                child = self.iri_to_name(a.sub)
                parent = self.iri_to_name(a.super)
                slot_isamap[child].add(parent)

            # domains become slot declarations
            if isinstance(a, ObjectPropertyDomain) or isinstance(a, DataPropertyDomain):
                if isinstance(a, ObjectPropertyDomain):
                    p = a.objectPropertyExpression.v
                else:
                    p = a.dataPropertyExpression.v
                sn = self.iri_to_name(p)
                dc = a.classExpression
                if isinstance(dc, Class):
                    c = self.iri_to_name(dc)
                    self.class_info(c, 'slots', sn, True)
                    #logging.error(f'Inferred {c} from domain of {p}')
                if isinstance(dc, ObjectUnionOf):
                    for x in dc.classExpressions:
                        if isinstance(x, Class):
                            c = self.iri_to_name(x)
                            self.class_info(c, 'slots', sn, True)

            if isinstance(a, ObjectPropertyRange):
                p = a.objectPropertyExpression.v
                sn = self.iri_to_name(p)
                rc = a.classExpression
                if isinstance(rc, Class):
                    self.slot_info(sn, 'range', self.iri_to_name(rc))

            if isinstance(a, DataPropertyRange):
                p = a.dataPropertyExpression.v
                sn = self.iri_to_name(p)
                rc = a.dataRange
                if isinstance(rc, Datatype):
                    logging.error('TODO')
                    #self.slot_info(sn, 'range', self.iri_to_name(rc))

            if isinstance(a, AnnotationPropertyRange):
                self.slot_info(self.iri_to_name(a.property),
                               'range', self.iri_to_name(a.range))


            if isinstance(a, Declaration):
                e = a.v
                uri_as_curie = str(e.v)
                if uri_as_curie.startswith(':'):
                    uri_as_curie = f'{name}{uri_as_curie}'
                if type(e) == Class:
                    cn = self.iri_to_name(e.v)
                    self.class_info(cn, 'class_uri', uri_as_curie)
                if type(e) in [ObjectProperty, DataProperty, AnnotationProperty]:
                    cn = self.iri_to_name(e.v)
                    self.slot_info(cn, 'slot_uri', uri_as_curie)


        for c, parents in isamap.items():
            parents = list(parents)
            p = parents.pop()
            self.class_info(c, 'is_a', p)
            for p in parents:
                self.class_info(c, 'mixins', p, True)
        for c, parents in slot_isamap.items():
            parents = list(parents)
            p = parents.pop()
            self.slot_info(c, 'is_a', p)
            for p in parents:
                self.slot_info(c, 'mixins', p, True)

        for a in ontology.axioms:
            if isinstance(a, AnnotationAssertion):
                p = a.property
                strp = str(p)
                sub = a.subject.v
                val = a.value.v
                if isinstance(sub, IRI):
                    sub = self.iri_to_name(sub)
                    if isinstance(val, Literal):
                        val = str(val.v)
                    elif isinstance(val, IRI):
                        val = val.v
                    else:
                        val = str(val)
                    if sub in classes:
                        t = 'classes'
                    elif sub in slots:
                        t = 'slots'
                    else:
                        logging.error(f'{sub} is not known')
                    if t is not None:
                        if strp == 'rdfs:comment':
                            self.element_info(t, sub, 'comments', val, multivalued=True)
                        elif strp == ':definition':
                            self.element_info(t, sub, 'description', val, multivalued=False)
                        elif strp == 'schema:rangeIncludes':
                            range_cn = self.iri_to_name(val)
                            logging.error(f'UNTESTED RANGE: schema.org {sub} {val} // {domain_cn}')
                            self.add_range(sub, range_cn)
                        elif strp == 'schema:domainIncludes':
                            domain_cn = self.iri_to_name(val)
                            logging.error(f'UNTESTED: schema.org {sub} {val} // {domain_cn}')
                            if domain_cn not in self.schema['classes']:
                                self.schema['classes'][domain_cn] = {}
                            if 'slots' not in self.schema['classes'][domain_cn]:
                                self.schema['classes'][domain_cn]['slots'] = []
                            self.schema['classes'][domain_cn]['slots'].append(sub)
                        else:
                            if self.include_unmapped_annotations:
                                self.element_info(t, sub, 'comments', f'{p} = {val}', multivalued=True)
        for cn, usage in slot_usage_map.items():
            schema_dict['classes'][cn]['slot_usage'] = usage
        for sn, s in schema_dict['slots'].items():
            if 'multivalued' not in s:
                s['multivalued'] = sn not in single_valued_slots
            if 'range' in s:
                if isinstance(s['range'], list):
                    rg = s['range']
                    if len(rg) == 0:
                        del s['range']
                    elif len(rg) == 1:
                        s['range'] = rg[0]
                    else:
                        del s['range']
                        s['any_of'] = [
                            {'range': x} for x in rg
                        ]
        if identifier is not None:
            slots[identifier] = {'identifier': True, 'range': 'uriorcurie'}
            for c in classes.values():
                if not c.get('is_a', None) and not c.get('mixins', []):
                    if 'slots' not in c:
                        c['slots'] = []
                    c['slots'].append(identifier)
        schema = SchemaDefinition(**schema_dict)
        return schema

    def class_info(self, *args, **kwargs):
        self.element_info('classes', *args, **kwargs)

    def slot_info(self, *args, **kwargs):
        self.element_info('slots', *args, **kwargs)

    def add_range(self, sn, range_cn):
        if sn not in self.schema['slots']:
            self.schema['slots'][sn] = {}
        if 'range' not in self.schema['slots'][sn]:
            self.schema['slots'][sn]['range'] = []
        self.schema['slots'][sn]['range'].append(range_cn)


    def element_info(self, type: str, cn: str, sn: str, v: Any, multivalued = False):
        if cn not in self.schema[type]:
            self.schema[type][cn] = {}
        c = self.schema[type][cn]
        if multivalued:
            if sn not in c:
                c[sn] = []
            c[sn].append(v)
        else:
            if sn in c and v != c[sn]:
                logging.error(f'Overwriting {sn} for {c} to {v}')
            c[sn] = v

    def iri_to_name(self, v):
        n = self._as_name(v)
        if n != v:
            self.mappings[n] = v
        return n

    def _as_name(self, v):
        v = str(v)
        for sep in ['#', '/', ':']:
            if sep in v:
                return v.split(sep)[-1]
        return v


@click.command()
@click.argument('owlfile')
@click.option('--name', '-n', help="Schema name")
@click.option('--identifier', '-I', help="Slot to use as identifier")
@click.option('--model-uri', help="Model URI prefix")
@click.option('--output', '-o', help="Path to saved yaml schema")
def owl2model(owlfile, output, **args):
    """
    Infer a model from OWL Ontology

    Note: input must be in functional syntax
    """
    sie = OwlImportEngine()
    schema = sie.convert(owlfile, **args)
    write_schema(schema, output)


if __name__ == '__main__':
    owl2model()





# Auto generated from dosdp_linkml.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-08-31 01:14
# Schema: DOSDP
#
# id: https://w3id.org/linkml/dosdp
# description:
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
DOSDP = CurieNamespace('dosdp', 'https://w3id.org/linkml/dosdp/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
DEFAULT_ = DOSDP


# Types

# Class references
class VarDeclarationVar(extended_str):
    pass


class AliasDeclarationAlias(extended_str):
    pass


@dataclass
class Pattern(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Pattern
    class_class_curie: ClassVar[str] = "dosdp:Pattern"
    class_name: ClassVar[str] = "Pattern"
    class_model_uri: ClassVar[URIRef] = DOSDP.Pattern

    pattern_name: Optional[str] = None
    pattern_iri: Optional[str] = None
    base_IRI: Optional[str] = None
    contributors: Optional[Union[str, List[str]]] = empty_list()
    description: Optional[str] = None
    examples: Optional[str] = None
    status: Optional[Union[str, "StatusOptions"]] = None
    tags: Optional[str] = None
    readable_identifiers: Optional[Union[str, List[str]]] = empty_list()
    classes: Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]] = empty_dict()
    objectProperties: Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]] = empty_dict()
    relations: Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]] = empty_dict()
    dataProperties: Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]] = empty_dict()
    annotationProperties: Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]] = empty_dict()
    vars: Optional[Union[Dict[Union[str, VarDeclarationVar], Union[dict, "VarDeclaration"]], List[Union[dict, "VarDeclaration"]]]] = empty_dict()
    list_vars: Optional[str] = None
    data_vars: Optional[str] = None
    data_list_vars: Optional[str] = None
    internal_vars: Optional[Union[str, List[str]]] = empty_list()
    substitutions: Optional[Union[Union[dict, "RegexSub"], List[Union[dict, "RegexSub"]]]] = empty_list()
    annotations: Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]] = empty_list()
    logical_axioms: Optional[Union[Union[dict, "PrintfOwl"], List[Union[dict, "PrintfOwl"]]]] = empty_list()
    equivalentTo: Optional[Union[dict, "PrintfOwlConvenience"]] = None
    subClassOf: Optional[Union[dict, "PrintfOwlConvenience"]] = None
    GCI: Optional[Union[dict, "PrintfOwlConvenience"]] = None
    disjointWith: Optional[Union[dict, "PrintfOwlConvenience"]] = None
    name: Optional[Union[dict, "PrintfAnnotationObo"]] = None
    comment: Optional[Union[dict, "PrintfAnnotationObo"]] = None
    definition: Optional[Union[dict, "PrintfAnnotationObo"]] = None
    namespace: Optional[Union[dict, "PrintfAnnotationObo"]] = None
    exact_synonym: Optional[Union[dict, "ListAnnotationObo"]] = None
    narrow_synonym: Optional[Union[dict, "ListAnnotationObo"]] = None
    related_synonym: Optional[Union[dict, "ListAnnotationObo"]] = None
    broad_synonym: Optional[Union[dict, "ListAnnotationObo"]] = None
    xref: Optional[Union[dict, "ListAnnotationObo"]] = None
    generated_synonyms: Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]] = empty_list()
    generated_narrow_synonyms: Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]] = empty_list()
    generated_broad_synonyms: Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]] = empty_list()
    generated_related_synonyms: Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]] = empty_list()
    instance_graph: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.pattern_name is not None and not isinstance(self.pattern_name, str):
            self.pattern_name = str(self.pattern_name)

        if self.pattern_iri is not None and not isinstance(self.pattern_iri, str):
            self.pattern_iri = str(self.pattern_iri)

        if self.base_IRI is not None and not isinstance(self.base_IRI, str):
            self.base_IRI = str(self.base_IRI)

        if not isinstance(self.contributors, list):
            self.contributors = [self.contributors] if self.contributors is not None else []
        self.contributors = [v if isinstance(v, str) else str(v) for v in self.contributors]

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.examples is not None and not isinstance(self.examples, str):
            self.examples = str(self.examples)

        if self.status is not None and not isinstance(self.status, StatusOptions):
            self.status = StatusOptions(self.status)

        if self.tags is not None and not isinstance(self.tags, str):
            self.tags = str(self.tags)

        if not isinstance(self.readable_identifiers, list):
            self.readable_identifiers = [self.readable_identifiers] if self.readable_identifiers is not None else []
        self.readable_identifiers = [v if isinstance(v, str) else str(v) for v in self.readable_identifiers]

        self._normalize_inlined_as_dict(slot_name="classes", slot_type=AliasDeclaration, key_name="alias", keyed=True)

        self._normalize_inlined_as_dict(slot_name="objectProperties", slot_type=AliasDeclaration, key_name="alias", keyed=True)

        self._normalize_inlined_as_dict(slot_name="relations", slot_type=AliasDeclaration, key_name="alias", keyed=True)

        self._normalize_inlined_as_dict(slot_name="dataProperties", slot_type=AliasDeclaration, key_name="alias", keyed=True)

        self._normalize_inlined_as_dict(slot_name="annotationProperties", slot_type=AliasDeclaration, key_name="alias", keyed=True)

        self._normalize_inlined_as_dict(slot_name="vars", slot_type=VarDeclaration, key_name="var", keyed=True)

        if self.list_vars is not None and not isinstance(self.list_vars, str):
            self.list_vars = str(self.list_vars)

        if self.data_vars is not None and not isinstance(self.data_vars, str):
            self.data_vars = str(self.data_vars)

        if self.data_list_vars is not None and not isinstance(self.data_list_vars, str):
            self.data_list_vars = str(self.data_list_vars)

        if not isinstance(self.internal_vars, list):
            self.internal_vars = [self.internal_vars] if self.internal_vars is not None else []
        self.internal_vars = [v if isinstance(v, str) else str(v) for v in self.internal_vars]

        self._normalize_inlined_as_dict(slot_name="substitutions", slot_type=RegexSub, key_name="out", keyed=False)

        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if not isinstance(self.logical_axioms, list):
            self.logical_axioms = [self.logical_axioms] if self.logical_axioms is not None else []
        self.logical_axioms = [v if isinstance(v, PrintfOwl) else PrintfOwl(**as_dict(v)) for v in self.logical_axioms]

        if self.equivalentTo is not None and not isinstance(self.equivalentTo, PrintfOwlConvenience):
            self.equivalentTo = PrintfOwlConvenience(**as_dict(self.equivalentTo))

        if self.subClassOf is not None and not isinstance(self.subClassOf, PrintfOwlConvenience):
            self.subClassOf = PrintfOwlConvenience(**as_dict(self.subClassOf))

        if self.GCI is not None and not isinstance(self.GCI, PrintfOwlConvenience):
            self.GCI = PrintfOwlConvenience(**as_dict(self.GCI))

        if self.disjointWith is not None and not isinstance(self.disjointWith, PrintfOwlConvenience):
            self.disjointWith = PrintfOwlConvenience(**as_dict(self.disjointWith))

        if self.name is not None and not isinstance(self.name, PrintfAnnotationObo):
            self.name = PrintfAnnotationObo(**as_dict(self.name))

        if self.comment is not None and not isinstance(self.comment, PrintfAnnotationObo):
            self.comment = PrintfAnnotationObo(**as_dict(self.comment))

        if self.definition is not None and not isinstance(self.definition, PrintfAnnotationObo):
            self.definition = PrintfAnnotationObo(**as_dict(self.definition))

        if self.namespace is not None and not isinstance(self.namespace, PrintfAnnotationObo):
            self.namespace = PrintfAnnotationObo(**as_dict(self.namespace))

        if self.exact_synonym is not None and not isinstance(self.exact_synonym, ListAnnotationObo):
            self.exact_synonym = ListAnnotationObo(**as_dict(self.exact_synonym))

        if self.narrow_synonym is not None and not isinstance(self.narrow_synonym, ListAnnotationObo):
            self.narrow_synonym = ListAnnotationObo(**as_dict(self.narrow_synonym))

        if self.related_synonym is not None and not isinstance(self.related_synonym, ListAnnotationObo):
            self.related_synonym = ListAnnotationObo(**as_dict(self.related_synonym))

        if self.broad_synonym is not None and not isinstance(self.broad_synonym, ListAnnotationObo):
            self.broad_synonym = ListAnnotationObo(**as_dict(self.broad_synonym))

        if self.xref is not None and not isinstance(self.xref, ListAnnotationObo):
            self.xref = ListAnnotationObo(**as_dict(self.xref))

        if not isinstance(self.generated_synonyms, list):
            self.generated_synonyms = [self.generated_synonyms] if self.generated_synonyms is not None else []
        self.generated_synonyms = [v if isinstance(v, PrintfAnnotationObo) else PrintfAnnotationObo(**as_dict(v)) for v in self.generated_synonyms]

        if not isinstance(self.generated_narrow_synonyms, list):
            self.generated_narrow_synonyms = [self.generated_narrow_synonyms] if self.generated_narrow_synonyms is not None else []
        self.generated_narrow_synonyms = [v if isinstance(v, PrintfAnnotationObo) else PrintfAnnotationObo(**as_dict(v)) for v in self.generated_narrow_synonyms]

        if not isinstance(self.generated_broad_synonyms, list):
            self.generated_broad_synonyms = [self.generated_broad_synonyms] if self.generated_broad_synonyms is not None else []
        self.generated_broad_synonyms = [v if isinstance(v, PrintfAnnotationObo) else PrintfAnnotationObo(**as_dict(v)) for v in self.generated_broad_synonyms]

        if not isinstance(self.generated_related_synonyms, list):
            self.generated_related_synonyms = [self.generated_related_synonyms] if self.generated_related_synonyms is not None else []
        self.generated_related_synonyms = [v if isinstance(v, PrintfAnnotationObo) else PrintfAnnotationObo(**as_dict(v)) for v in self.generated_related_synonyms]

        if self.instance_graph is not None and not isinstance(self.instance_graph, str):
            self.instance_graph = str(self.instance_graph)

        super().__post_init__(**kwargs)


class Declaration(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Declaration
    class_class_curie: ClassVar[str] = "dosdp:Declaration"
    class_name: ClassVar[str] = "Declaration"
    class_model_uri: ClassVar[URIRef] = DOSDP.Declaration


@dataclass
class VarDeclaration(Declaration):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.VarDeclaration
    class_class_curie: ClassVar[str] = "dosdp:VarDeclaration"
    class_name: ClassVar[str] = "VarDeclaration"
    class_model_uri: ClassVar[URIRef] = DOSDP.VarDeclaration

    var: Union[str, VarDeclarationVar] = None
    range: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.var):
            self.MissingRequiredField("var")
        if not isinstance(self.var, VarDeclarationVar):
            self.var = VarDeclarationVar(self.var)

        if self.range is not None and not isinstance(self.range, str):
            self.range = str(self.range)

        super().__post_init__(**kwargs)


@dataclass
class AliasDeclaration(Declaration):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.AliasDeclaration
    class_class_curie: ClassVar[str] = "dosdp:AliasDeclaration"
    class_name: ClassVar[str] = "AliasDeclaration"
    class_model_uri: ClassVar[URIRef] = DOSDP.AliasDeclaration

    alias: Union[str, AliasDeclarationAlias] = None
    curie: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.alias):
            self.MissingRequiredField("alias")
        if not isinstance(self.alias, AliasDeclarationAlias):
            self.alias = AliasDeclarationAlias(self.alias)

        if self.curie is not None and not isinstance(self.curie, str):
            self.curie = str(self.curie)

        super().__post_init__(**kwargs)


@dataclass
class MultiClausePrintf(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.MultiClausePrintf
    class_class_curie: ClassVar[str] = "dosdp:MultiClausePrintf"
    class_name: ClassVar[str] = "MultiClausePrintf"
    class_model_uri: ClassVar[URIRef] = DOSDP.MultiClausePrintf

    clauses: Union[Union[dict, "PrintfClause"], List[Union[dict, "PrintfClause"]]] = None
    sep: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.clauses):
            self.MissingRequiredField("clauses")
        self._normalize_inlined_as_dict(slot_name="clauses", slot_type=PrintfClause, key_name="text", keyed=False)

        if self.sep is not None and not isinstance(self.sep, str):
            self.sep = str(self.sep)

        super().__post_init__(**kwargs)


@dataclass
class PrintfClause(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.PrintfClause
    class_class_curie: ClassVar[str] = "dosdp:PrintfClause"
    class_name: ClassVar[str] = "PrintfClause"
    class_model_uri: ClassVar[URIRef] = DOSDP.PrintfClause

    text: str = None
    vars: Union[str, List[str]] = None
    sub_clauses: Optional[Union[Union[dict, MultiClausePrintf], List[Union[dict, MultiClausePrintf]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.text):
            self.MissingRequiredField("text")
        if not isinstance(self.text, str):
            self.text = str(self.text)

        if self._is_empty(self.vars):
            self.MissingRequiredField("vars")
        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        self._normalize_inlined_as_dict(slot_name="sub_clauses", slot_type=MultiClausePrintf, key_name="clauses", keyed=False)

        super().__post_init__(**kwargs)


class Function(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Function
    class_class_curie: ClassVar[str] = "dosdp:Function"
    class_name: ClassVar[str] = "Function"
    class_model_uri: ClassVar[URIRef] = DOSDP.Function


@dataclass
class Join(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Join
    class_class_curie: ClassVar[str] = "dosdp:Join"
    class_name: ClassVar[str] = "Join"
    class_model_uri: ClassVar[URIRef] = DOSDP.Join

    sep: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.sep is not None and not isinstance(self.sep, str):
            self.sep = str(self.sep)

        super().__post_init__(**kwargs)


@dataclass
class PrintfAnnotation(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.PrintfAnnotation
    class_class_curie: ClassVar[str] = "dosdp:PrintfAnnotation"
    class_name: ClassVar[str] = "PrintfAnnotation"
    class_model_uri: ClassVar[URIRef] = DOSDP.PrintfAnnotation

    annotationProperty: str = None
    annotations: Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]] = empty_list()
    text: Optional[str] = None
    vars: Optional[Union[str, List[str]]] = empty_list()
    multi_clause: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.annotationProperty):
            self.MissingRequiredField("annotationProperty")
        if not isinstance(self.annotationProperty, str):
            self.annotationProperty = str(self.annotationProperty)

        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        if self.multi_clause is not None and not isinstance(self.multi_clause, str):
            self.multi_clause = str(self.multi_clause)

        super().__post_init__(**kwargs)


@dataclass
class ListAnnotation(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.ListAnnotation
    class_class_curie: ClassVar[str] = "dosdp:ListAnnotation"
    class_name: ClassVar[str] = "ListAnnotation"
    class_model_uri: ClassVar[URIRef] = DOSDP.ListAnnotation

    annotationProperty: str = None
    value: str = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.annotationProperty):
            self.MissingRequiredField("annotationProperty")
        if not isinstance(self.annotationProperty, str):
            self.annotationProperty = str(self.annotationProperty)

        if self._is_empty(self.value):
            self.MissingRequiredField("value")
        if not isinstance(self.value, str):
            self.value = str(self.value)

        super().__post_init__(**kwargs)


@dataclass
class IriValueAnnotation(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.IriValueAnnotation
    class_class_curie: ClassVar[str] = "dosdp:IriValueAnnotation"
    class_name: ClassVar[str] = "IriValueAnnotation"
    class_model_uri: ClassVar[URIRef] = DOSDP.IriValueAnnotation

    annotationProperty: str = None
    var: str = None
    annotations: Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.annotationProperty):
            self.MissingRequiredField("annotationProperty")
        if not isinstance(self.annotationProperty, str):
            self.annotationProperty = str(self.annotationProperty)

        if self._is_empty(self.var):
            self.MissingRequiredField("var")
        if not isinstance(self.var, str):
            self.var = str(self.var)

        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        super().__post_init__(**kwargs)


class Annotation(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Annotation
    class_class_curie: ClassVar[str] = "dosdp:Annotation"
    class_name: ClassVar[str] = "Annotation"
    class_model_uri: ClassVar[URIRef] = DOSDP.Annotation


@dataclass
class Printf(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.Printf
    class_class_curie: ClassVar[str] = "dosdp:Printf"
    class_name: ClassVar[str] = "Printf"
    class_model_uri: ClassVar[URIRef] = DOSDP.Printf

    text: Optional[str] = None
    vars: Optional[Union[str, List[str]]] = empty_list()
    annotations: Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        super().__post_init__(**kwargs)


@dataclass
class PrintfOwl(Printf):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.PrintfOwl
    class_class_curie: ClassVar[str] = "dosdp:PrintfOwl"
    class_name: ClassVar[str] = "PrintfOwl"
    class_model_uri: ClassVar[URIRef] = DOSDP.PrintfOwl

    annotations: Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]] = empty_list()
    axiom_type: Optional[Union[str, "AxiomTypeOptions"]] = None
    text: Optional[str] = None
    vars: Optional[Union[str, List[str]]] = empty_list()
    multi_clause: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if self.axiom_type is not None and not isinstance(self.axiom_type, AxiomTypeOptions):
            self.axiom_type = AxiomTypeOptions(self.axiom_type)

        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        if self.multi_clause is not None and not isinstance(self.multi_clause, str):
            self.multi_clause = str(self.multi_clause)

        super().__post_init__(**kwargs)


@dataclass
class PrintfOwlConvenience(Printf):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.PrintfOwlConvenience
    class_class_curie: ClassVar[str] = "dosdp:PrintfOwlConvenience"
    class_name: ClassVar[str] = "PrintfOwlConvenience"
    class_model_uri: ClassVar[URIRef] = DOSDP.PrintfOwlConvenience

    annotations: Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]] = empty_list()
    text: Optional[str] = None
    vars: Optional[Union[str, List[str]]] = empty_list()
    multi_clause: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        if self.multi_clause is not None and not isinstance(self.multi_clause, str):
            self.multi_clause = str(self.multi_clause)

        super().__post_init__(**kwargs)


@dataclass
class RegexSub(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.RegexSub
    class_class_curie: ClassVar[str] = "dosdp:RegexSub"
    class_name: ClassVar[str] = "RegexSub"
    class_model_uri: ClassVar[URIRef] = DOSDP.RegexSub

    out: str = None
    match: str = None
    sub: str = None
    _in: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.out):
            self.MissingRequiredField("out")
        if not isinstance(self.out, str):
            self.out = str(self.out)

        if self._is_empty(self.match):
            self.MissingRequiredField("match")
        if not isinstance(self.match, str):
            self.match = str(self.match)

        if self._is_empty(self.sub):
            self.MissingRequiredField("sub")
        if not isinstance(self.sub, str):
            self.sub = str(self.sub)

        if self._in is not None and not isinstance(self._in, str):
            self._in = str(self._in)

        super().__post_init__(**kwargs)


@dataclass
class OPA(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.OPA
    class_class_curie: ClassVar[str] = "dosdp:OPA"
    class_name: ClassVar[str] = "OPA"
    class_model_uri: ClassVar[URIRef] = DOSDP.OPA

    edge: Union[str, List[str]] = None
    annotations: Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]] = empty_list()
    _not: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.edge):
            self.MissingRequiredField("edge")
        if not isinstance(self.edge, list):
            self.edge = [self.edge] if self.edge is not None else []
        self.edge = [v if isinstance(v, str) else str(v) for v in self.edge]

        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if self._not is not None and not isinstance(self._not, Bool):
            self._not = Bool(self._not)

        super().__post_init__(**kwargs)


@dataclass
class PrintfAnnotationObo(Printf):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.PrintfAnnotationObo
    class_class_curie: ClassVar[str] = "dosdp:PrintfAnnotationObo"
    class_name: ClassVar[str] = "PrintfAnnotationObo"
    class_model_uri: ClassVar[URIRef] = DOSDP.PrintfAnnotationObo

    annotations: Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]] = empty_list()
    xrefs: Optional[str] = None
    text: Optional[str] = None
    vars: Optional[Union[str, List[str]]] = empty_list()
    multi_clause: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.annotations, list):
            self.annotations = [self.annotations] if self.annotations is not None else []
        self.annotations = [v if isinstance(v, Annotation) else Annotation(**as_dict(v)) for v in self.annotations]

        if self.xrefs is not None and not isinstance(self.xrefs, str):
            self.xrefs = str(self.xrefs)

        if self.text is not None and not isinstance(self.text, str):
            self.text = str(self.text)

        if not isinstance(self.vars, list):
            self.vars = [self.vars] if self.vars is not None else []
        self.vars = [v if isinstance(v, str) else str(v) for v in self.vars]

        if self.multi_clause is not None and not isinstance(self.multi_clause, str):
            self.multi_clause = str(self.multi_clause)

        super().__post_init__(**kwargs)


@dataclass
class ListAnnotationObo(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = DOSDP.ListAnnotationObo
    class_class_curie: ClassVar[str] = "dosdp:ListAnnotationObo"
    class_name: ClassVar[str] = "ListAnnotationObo"
    class_model_uri: ClassVar[URIRef] = DOSDP.ListAnnotationObo

    value: str = None
    xrefs: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.value):
            self.MissingRequiredField("value")
        if not isinstance(self.value, str):
            self.value = str(self.value)

        if self.xrefs is not None and not isinstance(self.xrefs, str):
            self.xrefs = str(self.xrefs)

        super().__post_init__(**kwargs)


# Enumerations
class AxiomTypeOptions(EnumDefinitionImpl):

    equivalentTo = PermissibleValue(text="equivalentTo")
    subClassOf = PermissibleValue(text="subClassOf")
    disjointWith = PermissibleValue(text="disjointWith")
    GCI = PermissibleValue(text="GCI")

    _defn = EnumDefinition(
        name="AxiomTypeOptions",
    )

class StatusOptions(EnumDefinitionImpl):

    development = PermissibleValue(text="development")
    published = PermissibleValue(text="published")

    _defn = EnumDefinition(
        name="StatusOptions",
    )

# Slots
class slots:
    pass

slots.sep = Slot(uri=DOSDP.sep, name="sep", curie=DOSDP.curie('sep'),
                   model_uri=DOSDP.sep, domain=None, range=Optional[str])

slots.clauses = Slot(uri=DOSDP.clauses, name="clauses", curie=DOSDP.curie('clauses'),
                   model_uri=DOSDP.clauses, domain=None, range=Union[Union[dict, PrintfClause], List[Union[dict, PrintfClause]]])

slots.text = Slot(uri=DOSDP.text, name="text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.text, domain=None, range=Optional[str])

slots.vars = Slot(uri=DOSDP.vars, name="vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.vars, domain=None, range=Optional[Union[str, List[str]]])

slots.sub_clauses = Slot(uri=DOSDP.sub_clauses, name="sub_clauses", curie=DOSDP.curie('sub_clauses'),
                   model_uri=DOSDP.sub_clauses, domain=None, range=Optional[Union[Union[dict, MultiClausePrintf], List[Union[dict, MultiClausePrintf]]]])

slots.annotationProperty = Slot(uri=DOSDP.annotationProperty, name="annotationProperty", curie=DOSDP.curie('annotationProperty'),
                   model_uri=DOSDP.annotationProperty, domain=None, range=str)

slots.annotations = Slot(uri=DOSDP.annotations, name="annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.annotations, domain=None, range=Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]])

slots.multi_clause = Slot(uri=DOSDP.multi_clause, name="multi_clause", curie=DOSDP.curie('multi_clause'),
                   model_uri=DOSDP.multi_clause, domain=None, range=Optional[str])

slots.value = Slot(uri=DOSDP.value, name="value", curie=DOSDP.curie('value'),
                   model_uri=DOSDP.value, domain=None, range=str)

slots.var = Slot(uri=DOSDP.var, name="var", curie=DOSDP.curie('var'),
                   model_uri=DOSDP.var, domain=None, range=str)

slots.axiom_type = Slot(uri=DOSDP.axiom_type, name="axiom_type", curie=DOSDP.curie('axiom_type'),
                   model_uri=DOSDP.axiom_type, domain=None, range=Optional[Union[str, "AxiomTypeOptions"]])

slots._in = Slot(uri=DOSDP._in, name="_in", curie=DOSDP.curie('_in'),
                   model_uri=DOSDP._in, domain=None, range=Optional[str])

slots.out = Slot(uri=DOSDP.out, name="out", curie=DOSDP.curie('out'),
                   model_uri=DOSDP.out, domain=None, range=str)

slots.match = Slot(uri=DOSDP.match, name="match", curie=DOSDP.curie('match'),
                   model_uri=DOSDP.match, domain=None, range=str)

slots.sub = Slot(uri=DOSDP.sub, name="sub", curie=DOSDP.curie('sub'),
                   model_uri=DOSDP.sub, domain=None, range=str)

slots.edge = Slot(uri=DOSDP.edge, name="edge", curie=DOSDP.curie('edge'),
                   model_uri=DOSDP.edge, domain=None, range=Union[str, List[str]])

slots._not = Slot(uri=DOSDP._not, name="_not", curie=DOSDP.curie('_not'),
                   model_uri=DOSDP._not, domain=None, range=Optional[Union[bool, Bool]])

slots.xrefs = Slot(uri=DOSDP.xrefs, name="xrefs", curie=DOSDP.curie('xrefs'),
                   model_uri=DOSDP.xrefs, domain=None, range=Optional[str])

slots.pattern_name = Slot(uri=DOSDP.pattern_name, name="pattern_name", curie=DOSDP.curie('pattern_name'),
                   model_uri=DOSDP.pattern_name, domain=None, range=Optional[str])

slots.pattern_iri = Slot(uri=DOSDP.pattern_iri, name="pattern_iri", curie=DOSDP.curie('pattern_iri'),
                   model_uri=DOSDP.pattern_iri, domain=None, range=Optional[str])

slots.base_IRI = Slot(uri=DOSDP.base_IRI, name="base_IRI", curie=DOSDP.curie('base_IRI'),
                   model_uri=DOSDP.base_IRI, domain=None, range=Optional[str])

slots.contributors = Slot(uri=DOSDP.contributors, name="contributors", curie=DOSDP.curie('contributors'),
                   model_uri=DOSDP.contributors, domain=None, range=Optional[Union[str, List[str]]])

slots.description = Slot(uri=DOSDP.description, name="description", curie=DOSDP.curie('description'),
                   model_uri=DOSDP.description, domain=None, range=Optional[str])

slots.examples = Slot(uri=DOSDP.examples, name="examples", curie=DOSDP.curie('examples'),
                   model_uri=DOSDP.examples, domain=None, range=Optional[str])

slots.status = Slot(uri=DOSDP.status, name="status", curie=DOSDP.curie('status'),
                   model_uri=DOSDP.status, domain=None, range=Optional[Union[str, "StatusOptions"]])

slots.tags = Slot(uri=DOSDP.tags, name="tags", curie=DOSDP.curie('tags'),
                   model_uri=DOSDP.tags, domain=None, range=Optional[str])

slots.readable_identifiers = Slot(uri=DOSDP.readable_identifiers, name="readable_identifiers", curie=DOSDP.curie('readable_identifiers'),
                   model_uri=DOSDP.readable_identifiers, domain=None, range=Optional[Union[str, List[str]]])

slots.classes = Slot(uri=DOSDP.classes, name="classes", curie=DOSDP.curie('classes'),
                   model_uri=DOSDP.classes, domain=None, range=Optional[str])

slots.objectProperties = Slot(uri=DOSDP.objectProperties, name="objectProperties", curie=DOSDP.curie('objectProperties'),
                   model_uri=DOSDP.objectProperties, domain=None, range=Optional[str])

slots.relations = Slot(uri=DOSDP.relations, name="relations", curie=DOSDP.curie('relations'),
                   model_uri=DOSDP.relations, domain=None, range=Optional[str])

slots.dataProperties = Slot(uri=DOSDP.dataProperties, name="dataProperties", curie=DOSDP.curie('dataProperties'),
                   model_uri=DOSDP.dataProperties, domain=None, range=Optional[str])

slots.annotationProperties = Slot(uri=DOSDP.annotationProperties, name="annotationProperties", curie=DOSDP.curie('annotationProperties'),
                   model_uri=DOSDP.annotationProperties, domain=None, range=Optional[str])

slots.list_vars = Slot(uri=DOSDP.list_vars, name="list_vars", curie=DOSDP.curie('list_vars'),
                   model_uri=DOSDP.list_vars, domain=None, range=Optional[str])

slots.data_vars = Slot(uri=DOSDP.data_vars, name="data_vars", curie=DOSDP.curie('data_vars'),
                   model_uri=DOSDP.data_vars, domain=None, range=Optional[str])

slots.data_list_vars = Slot(uri=DOSDP.data_list_vars, name="data_list_vars", curie=DOSDP.curie('data_list_vars'),
                   model_uri=DOSDP.data_list_vars, domain=None, range=Optional[str])

slots.internal_vars = Slot(uri=DOSDP.internal_vars, name="internal_vars", curie=DOSDP.curie('internal_vars'),
                   model_uri=DOSDP.internal_vars, domain=None, range=Optional[Union[str, List[str]]])

slots.substitutions = Slot(uri=DOSDP.substitutions, name="substitutions", curie=DOSDP.curie('substitutions'),
                   model_uri=DOSDP.substitutions, domain=None, range=Optional[Union[Union[dict, RegexSub], List[Union[dict, RegexSub]]]])

slots.logical_axioms = Slot(uri=DOSDP.logical_axioms, name="logical_axioms", curie=DOSDP.curie('logical_axioms'),
                   model_uri=DOSDP.logical_axioms, domain=None, range=Optional[Union[Union[dict, PrintfOwl], List[Union[dict, PrintfOwl]]]])

slots.equivalentTo = Slot(uri=DOSDP.equivalentTo, name="equivalentTo", curie=DOSDP.curie('equivalentTo'),
                   model_uri=DOSDP.equivalentTo, domain=None, range=Optional[Union[dict, PrintfOwlConvenience]])

slots.subClassOf = Slot(uri=DOSDP.subClassOf, name="subClassOf", curie=DOSDP.curie('subClassOf'),
                   model_uri=DOSDP.subClassOf, domain=None, range=Optional[Union[dict, PrintfOwlConvenience]])

slots.GCI = Slot(uri=DOSDP.GCI, name="GCI", curie=DOSDP.curie('GCI'),
                   model_uri=DOSDP.GCI, domain=None, range=Optional[Union[dict, PrintfOwlConvenience]])

slots.disjointWith = Slot(uri=DOSDP.disjointWith, name="disjointWith", curie=DOSDP.curie('disjointWith'),
                   model_uri=DOSDP.disjointWith, domain=None, range=Optional[Union[dict, PrintfOwlConvenience]])

slots.name = Slot(uri=DOSDP.name, name="name", curie=DOSDP.curie('name'),
                   model_uri=DOSDP.name, domain=None, range=Optional[Union[dict, PrintfAnnotationObo]])

slots.comment = Slot(uri=DOSDP.comment, name="comment", curie=DOSDP.curie('comment'),
                   model_uri=DOSDP.comment, domain=None, range=Optional[Union[dict, PrintfAnnotationObo]])

slots.definition = Slot(uri=DOSDP.definition, name="definition", curie=DOSDP.curie('definition'),
                   model_uri=DOSDP.definition, domain=None, range=Optional[Union[dict, PrintfAnnotationObo]])

slots.namespace = Slot(uri=DOSDP.namespace, name="namespace", curie=DOSDP.curie('namespace'),
                   model_uri=DOSDP.namespace, domain=None, range=Optional[Union[dict, PrintfAnnotationObo]])

slots.exact_synonym = Slot(uri=DOSDP.exact_synonym, name="exact_synonym", curie=DOSDP.curie('exact_synonym'),
                   model_uri=DOSDP.exact_synonym, domain=None, range=Optional[Union[dict, ListAnnotationObo]])

slots.narrow_synonym = Slot(uri=DOSDP.narrow_synonym, name="narrow_synonym", curie=DOSDP.curie('narrow_synonym'),
                   model_uri=DOSDP.narrow_synonym, domain=None, range=Optional[Union[dict, ListAnnotationObo]])

slots.related_synonym = Slot(uri=DOSDP.related_synonym, name="related_synonym", curie=DOSDP.curie('related_synonym'),
                   model_uri=DOSDP.related_synonym, domain=None, range=Optional[Union[dict, ListAnnotationObo]])

slots.broad_synonym = Slot(uri=DOSDP.broad_synonym, name="broad_synonym", curie=DOSDP.curie('broad_synonym'),
                   model_uri=DOSDP.broad_synonym, domain=None, range=Optional[Union[dict, ListAnnotationObo]])

slots.xref = Slot(uri=DOSDP.xref, name="xref", curie=DOSDP.curie('xref'),
                   model_uri=DOSDP.xref, domain=None, range=Optional[Union[dict, ListAnnotationObo]])

slots.generated_synonyms = Slot(uri=DOSDP.generated_synonyms, name="generated_synonyms", curie=DOSDP.curie('generated_synonyms'),
                   model_uri=DOSDP.generated_synonyms, domain=None, range=Optional[Union[Union[dict, PrintfAnnotationObo], List[Union[dict, PrintfAnnotationObo]]]])

slots.generated_narrow_synonyms = Slot(uri=DOSDP.generated_narrow_synonyms, name="generated_narrow_synonyms", curie=DOSDP.curie('generated_narrow_synonyms'),
                   model_uri=DOSDP.generated_narrow_synonyms, domain=None, range=Optional[Union[Union[dict, PrintfAnnotationObo], List[Union[dict, PrintfAnnotationObo]]]])

slots.generated_broad_synonyms = Slot(uri=DOSDP.generated_broad_synonyms, name="generated_broad_synonyms", curie=DOSDP.curie('generated_broad_synonyms'),
                   model_uri=DOSDP.generated_broad_synonyms, domain=None, range=Optional[Union[Union[dict, PrintfAnnotationObo], List[Union[dict, PrintfAnnotationObo]]]])

slots.generated_related_synonyms = Slot(uri=DOSDP.generated_related_synonyms, name="generated_related_synonyms", curie=DOSDP.curie('generated_related_synonyms'),
                   model_uri=DOSDP.generated_related_synonyms, domain=None, range=Optional[Union[Union[dict, PrintfAnnotationObo], List[Union[dict, PrintfAnnotationObo]]]])

slots.instance_graph = Slot(uri=DOSDP.instance_graph, name="instance_graph", curie=DOSDP.curie('instance_graph'),
                   model_uri=DOSDP.instance_graph, domain=None, range=Optional[str])

slots.varDeclaration__var = Slot(uri=DOSDP.var, name="varDeclaration__var", curie=DOSDP.curie('var'),
                   model_uri=DOSDP.varDeclaration__var, domain=None, range=URIRef)

slots.varDeclaration__range = Slot(uri=DOSDP.range, name="varDeclaration__range", curie=DOSDP.curie('range'),
                   model_uri=DOSDP.varDeclaration__range, domain=None, range=Optional[str])

slots.aliasDeclaration__alias = Slot(uri=DOSDP.alias, name="aliasDeclaration__alias", curie=DOSDP.curie('alias'),
                   model_uri=DOSDP.aliasDeclaration__alias, domain=None, range=URIRef)

slots.aliasDeclaration__curie = Slot(uri=DOSDP.curie, name="aliasDeclaration__curie", curie=DOSDP.curie('curie'),
                   model_uri=DOSDP.aliasDeclaration__curie, domain=None, range=Optional[str])

slots.Pattern_pattern_name = Slot(uri=DOSDP.pattern_name, name="Pattern_pattern_name", curie=DOSDP.curie('pattern_name'),
                   model_uri=DOSDP.Pattern_pattern_name, domain=Pattern, range=Optional[str])

slots.Pattern_pattern_iri = Slot(uri=DOSDP.pattern_iri, name="Pattern_pattern_iri", curie=DOSDP.curie('pattern_iri'),
                   model_uri=DOSDP.Pattern_pattern_iri, domain=Pattern, range=Optional[str])

slots.Pattern_base_IRI = Slot(uri=DOSDP.base_IRI, name="Pattern_base_IRI", curie=DOSDP.curie('base_IRI'),
                   model_uri=DOSDP.Pattern_base_IRI, domain=Pattern, range=Optional[str])

slots.Pattern_contributors = Slot(uri=DOSDP.contributors, name="Pattern_contributors", curie=DOSDP.curie('contributors'),
                   model_uri=DOSDP.Pattern_contributors, domain=Pattern, range=Optional[Union[str, List[str]]])

slots.Pattern_description = Slot(uri=DOSDP.description, name="Pattern_description", curie=DOSDP.curie('description'),
                   model_uri=DOSDP.Pattern_description, domain=Pattern, range=Optional[str])

slots.Pattern_examples = Slot(uri=DOSDP.examples, name="Pattern_examples", curie=DOSDP.curie('examples'),
                   model_uri=DOSDP.Pattern_examples, domain=Pattern, range=Optional[str])

slots.Pattern_status = Slot(uri=DOSDP.status, name="Pattern_status", curie=DOSDP.curie('status'),
                   model_uri=DOSDP.Pattern_status, domain=Pattern, range=Optional[Union[str, "StatusOptions"]])

slots.Pattern_tags = Slot(uri=DOSDP.tags, name="Pattern_tags", curie=DOSDP.curie('tags'),
                   model_uri=DOSDP.Pattern_tags, domain=Pattern, range=Optional[str])

slots.Pattern_readable_identifiers = Slot(uri=DOSDP.readable_identifiers, name="Pattern_readable_identifiers", curie=DOSDP.curie('readable_identifiers'),
                   model_uri=DOSDP.Pattern_readable_identifiers, domain=Pattern, range=Optional[Union[str, List[str]]])

slots.Pattern_classes = Slot(uri=DOSDP.classes, name="Pattern_classes", curie=DOSDP.curie('classes'),
                   model_uri=DOSDP.Pattern_classes, domain=Pattern, range=Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]])

slots.Pattern_objectProperties = Slot(uri=DOSDP.objectProperties, name="Pattern_objectProperties", curie=DOSDP.curie('objectProperties'),
                   model_uri=DOSDP.Pattern_objectProperties, domain=Pattern, range=Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]])

slots.Pattern_relations = Slot(uri=DOSDP.relations, name="Pattern_relations", curie=DOSDP.curie('relations'),
                   model_uri=DOSDP.Pattern_relations, domain=Pattern, range=Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]])

slots.Pattern_dataProperties = Slot(uri=DOSDP.dataProperties, name="Pattern_dataProperties", curie=DOSDP.curie('dataProperties'),
                   model_uri=DOSDP.Pattern_dataProperties, domain=Pattern, range=Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]])

slots.Pattern_annotationProperties = Slot(uri=DOSDP.annotationProperties, name="Pattern_annotationProperties", curie=DOSDP.curie('annotationProperties'),
                   model_uri=DOSDP.Pattern_annotationProperties, domain=Pattern, range=Optional[Union[Dict[Union[str, AliasDeclarationAlias], Union[dict, "AliasDeclaration"]], List[Union[dict, "AliasDeclaration"]]]])

slots.Pattern_vars = Slot(uri=DOSDP.vars, name="Pattern_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.Pattern_vars, domain=Pattern, range=Optional[Union[Dict[Union[str, VarDeclarationVar], Union[dict, "VarDeclaration"]], List[Union[dict, "VarDeclaration"]]]])

slots.Pattern_list_vars = Slot(uri=DOSDP.list_vars, name="Pattern_list_vars", curie=DOSDP.curie('list_vars'),
                   model_uri=DOSDP.Pattern_list_vars, domain=Pattern, range=Optional[str])

slots.Pattern_data_vars = Slot(uri=DOSDP.data_vars, name="Pattern_data_vars", curie=DOSDP.curie('data_vars'),
                   model_uri=DOSDP.Pattern_data_vars, domain=Pattern, range=Optional[str])

slots.Pattern_data_list_vars = Slot(uri=DOSDP.data_list_vars, name="Pattern_data_list_vars", curie=DOSDP.curie('data_list_vars'),
                   model_uri=DOSDP.Pattern_data_list_vars, domain=Pattern, range=Optional[str])

slots.Pattern_internal_vars = Slot(uri=DOSDP.internal_vars, name="Pattern_internal_vars", curie=DOSDP.curie('internal_vars'),
                   model_uri=DOSDP.Pattern_internal_vars, domain=Pattern, range=Optional[Union[str, List[str]]])

slots.Pattern_substitutions = Slot(uri=DOSDP.substitutions, name="Pattern_substitutions", curie=DOSDP.curie('substitutions'),
                   model_uri=DOSDP.Pattern_substitutions, domain=Pattern, range=Optional[Union[Union[dict, "RegexSub"], List[Union[dict, "RegexSub"]]]])

slots.Pattern_annotations = Slot(uri=DOSDP.annotations, name="Pattern_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.Pattern_annotations, domain=Pattern, range=Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]])

slots.Pattern_logical_axioms = Slot(uri=DOSDP.logical_axioms, name="Pattern_logical_axioms", curie=DOSDP.curie('logical_axioms'),
                   model_uri=DOSDP.Pattern_logical_axioms, domain=Pattern, range=Optional[Union[Union[dict, "PrintfOwl"], List[Union[dict, "PrintfOwl"]]]])

slots.Pattern_equivalentTo = Slot(uri=DOSDP.equivalentTo, name="Pattern_equivalentTo", curie=DOSDP.curie('equivalentTo'),
                   model_uri=DOSDP.Pattern_equivalentTo, domain=Pattern, range=Optional[Union[dict, "PrintfOwlConvenience"]])

slots.Pattern_subClassOf = Slot(uri=DOSDP.subClassOf, name="Pattern_subClassOf", curie=DOSDP.curie('subClassOf'),
                   model_uri=DOSDP.Pattern_subClassOf, domain=Pattern, range=Optional[Union[dict, "PrintfOwlConvenience"]])

slots.Pattern_GCI = Slot(uri=DOSDP.GCI, name="Pattern_GCI", curie=DOSDP.curie('GCI'),
                   model_uri=DOSDP.Pattern_GCI, domain=Pattern, range=Optional[Union[dict, "PrintfOwlConvenience"]])

slots.Pattern_disjointWith = Slot(uri=DOSDP.disjointWith, name="Pattern_disjointWith", curie=DOSDP.curie('disjointWith'),
                   model_uri=DOSDP.Pattern_disjointWith, domain=Pattern, range=Optional[Union[dict, "PrintfOwlConvenience"]])

slots.Pattern_name = Slot(uri=DOSDP.name, name="Pattern_name", curie=DOSDP.curie('name'),
                   model_uri=DOSDP.Pattern_name, domain=Pattern, range=Optional[Union[dict, "PrintfAnnotationObo"]])

slots.Pattern_comment = Slot(uri=DOSDP.comment, name="Pattern_comment", curie=DOSDP.curie('comment'),
                   model_uri=DOSDP.Pattern_comment, domain=Pattern, range=Optional[Union[dict, "PrintfAnnotationObo"]])

slots.Pattern_definition = Slot(uri=DOSDP.definition, name="Pattern_definition", curie=DOSDP.curie('definition'),
                   model_uri=DOSDP.Pattern_definition, domain=Pattern, range=Optional[Union[dict, "PrintfAnnotationObo"]])

slots.Pattern_namespace = Slot(uri=DOSDP.namespace, name="Pattern_namespace", curie=DOSDP.curie('namespace'),
                   model_uri=DOSDP.Pattern_namespace, domain=Pattern, range=Optional[Union[dict, "PrintfAnnotationObo"]])

slots.Pattern_exact_synonym = Slot(uri=DOSDP.exact_synonym, name="Pattern_exact_synonym", curie=DOSDP.curie('exact_synonym'),
                   model_uri=DOSDP.Pattern_exact_synonym, domain=Pattern, range=Optional[Union[dict, "ListAnnotationObo"]])

slots.Pattern_narrow_synonym = Slot(uri=DOSDP.narrow_synonym, name="Pattern_narrow_synonym", curie=DOSDP.curie('narrow_synonym'),
                   model_uri=DOSDP.Pattern_narrow_synonym, domain=Pattern, range=Optional[Union[dict, "ListAnnotationObo"]])

slots.Pattern_related_synonym = Slot(uri=DOSDP.related_synonym, name="Pattern_related_synonym", curie=DOSDP.curie('related_synonym'),
                   model_uri=DOSDP.Pattern_related_synonym, domain=Pattern, range=Optional[Union[dict, "ListAnnotationObo"]])

slots.Pattern_broad_synonym = Slot(uri=DOSDP.broad_synonym, name="Pattern_broad_synonym", curie=DOSDP.curie('broad_synonym'),
                   model_uri=DOSDP.Pattern_broad_synonym, domain=Pattern, range=Optional[Union[dict, "ListAnnotationObo"]])

slots.Pattern_xref = Slot(uri=DOSDP.xref, name="Pattern_xref", curie=DOSDP.curie('xref'),
                   model_uri=DOSDP.Pattern_xref, domain=Pattern, range=Optional[Union[dict, "ListAnnotationObo"]])

slots.Pattern_generated_synonyms = Slot(uri=DOSDP.generated_synonyms, name="Pattern_generated_synonyms", curie=DOSDP.curie('generated_synonyms'),
                   model_uri=DOSDP.Pattern_generated_synonyms, domain=Pattern, range=Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]])

slots.Pattern_generated_narrow_synonyms = Slot(uri=DOSDP.generated_narrow_synonyms, name="Pattern_generated_narrow_synonyms", curie=DOSDP.curie('generated_narrow_synonyms'),
                   model_uri=DOSDP.Pattern_generated_narrow_synonyms, domain=Pattern, range=Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]])

slots.Pattern_generated_broad_synonyms = Slot(uri=DOSDP.generated_broad_synonyms, name="Pattern_generated_broad_synonyms", curie=DOSDP.curie('generated_broad_synonyms'),
                   model_uri=DOSDP.Pattern_generated_broad_synonyms, domain=Pattern, range=Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]])

slots.Pattern_generated_related_synonyms = Slot(uri=DOSDP.generated_related_synonyms, name="Pattern_generated_related_synonyms", curie=DOSDP.curie('generated_related_synonyms'),
                   model_uri=DOSDP.Pattern_generated_related_synonyms, domain=Pattern, range=Optional[Union[Union[dict, "PrintfAnnotationObo"], List[Union[dict, "PrintfAnnotationObo"]]]])

slots.Pattern_instance_graph = Slot(uri=DOSDP.instance_graph, name="Pattern_instance_graph", curie=DOSDP.curie('instance_graph'),
                   model_uri=DOSDP.Pattern_instance_graph, domain=Pattern, range=Optional[str])

slots.MultiClausePrintf_sep = Slot(uri=DOSDP.sep, name="MultiClausePrintf_sep", curie=DOSDP.curie('sep'),
                   model_uri=DOSDP.MultiClausePrintf_sep, domain=MultiClausePrintf, range=Optional[str])

slots.MultiClausePrintf_clauses = Slot(uri=DOSDP.clauses, name="MultiClausePrintf_clauses", curie=DOSDP.curie('clauses'),
                   model_uri=DOSDP.MultiClausePrintf_clauses, domain=MultiClausePrintf, range=Union[Union[dict, "PrintfClause"], List[Union[dict, "PrintfClause"]]])

slots.PrintfClause_text = Slot(uri=DOSDP.text, name="PrintfClause_text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.PrintfClause_text, domain=PrintfClause, range=str)

slots.PrintfClause_vars = Slot(uri=DOSDP.vars, name="PrintfClause_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.PrintfClause_vars, domain=PrintfClause, range=Union[str, List[str]])

slots.PrintfClause_sub_clauses = Slot(uri=DOSDP.sub_clauses, name="PrintfClause_sub_clauses", curie=DOSDP.curie('sub_clauses'),
                   model_uri=DOSDP.PrintfClause_sub_clauses, domain=PrintfClause, range=Optional[Union[Union[dict, MultiClausePrintf], List[Union[dict, MultiClausePrintf]]]])

slots.Join_sep = Slot(uri=DOSDP.sep, name="Join_sep", curie=DOSDP.curie('sep'),
                   model_uri=DOSDP.Join_sep, domain=Join, range=Optional[str])

slots.PrintfAnnotation_annotationProperty = Slot(uri=DOSDP.annotationProperty, name="PrintfAnnotation_annotationProperty", curie=DOSDP.curie('annotationProperty'),
                   model_uri=DOSDP.PrintfAnnotation_annotationProperty, domain=PrintfAnnotation, range=str)

slots.PrintfAnnotation_annotations = Slot(uri=DOSDP.annotations, name="PrintfAnnotation_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.PrintfAnnotation_annotations, domain=PrintfAnnotation, range=Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]])

slots.PrintfAnnotation_text = Slot(uri=DOSDP.text, name="PrintfAnnotation_text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.PrintfAnnotation_text, domain=PrintfAnnotation, range=Optional[str])

slots.PrintfAnnotation_vars = Slot(uri=DOSDP.vars, name="PrintfAnnotation_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.PrintfAnnotation_vars, domain=PrintfAnnotation, range=Optional[Union[str, List[str]]])

slots.PrintfAnnotation_multi_clause = Slot(uri=DOSDP.multi_clause, name="PrintfAnnotation_multi_clause", curie=DOSDP.curie('multi_clause'),
                   model_uri=DOSDP.PrintfAnnotation_multi_clause, domain=PrintfAnnotation, range=Optional[str])

slots.ListAnnotation_annotationProperty = Slot(uri=DOSDP.annotationProperty, name="ListAnnotation_annotationProperty", curie=DOSDP.curie('annotationProperty'),
                   model_uri=DOSDP.ListAnnotation_annotationProperty, domain=ListAnnotation, range=str)

slots.ListAnnotation_value = Slot(uri=DOSDP.value, name="ListAnnotation_value", curie=DOSDP.curie('value'),
                   model_uri=DOSDP.ListAnnotation_value, domain=ListAnnotation, range=str)

slots.IriValueAnnotation_annotationProperty = Slot(uri=DOSDP.annotationProperty, name="IriValueAnnotation_annotationProperty", curie=DOSDP.curie('annotationProperty'),
                   model_uri=DOSDP.IriValueAnnotation_annotationProperty, domain=IriValueAnnotation, range=str)

slots.IriValueAnnotation_var = Slot(uri=DOSDP.var, name="IriValueAnnotation_var", curie=DOSDP.curie('var'),
                   model_uri=DOSDP.IriValueAnnotation_var, domain=IriValueAnnotation, range=str)

slots.IriValueAnnotation_annotations = Slot(uri=DOSDP.annotations, name="IriValueAnnotation_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.IriValueAnnotation_annotations, domain=IriValueAnnotation, range=Optional[Union[Union[dict, "Annotation"], List[Union[dict, "Annotation"]]]])

slots.PrintfOwl_annotations = Slot(uri=DOSDP.annotations, name="PrintfOwl_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.PrintfOwl_annotations, domain=PrintfOwl, range=Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]])

slots.PrintfOwl_axiom_type = Slot(uri=DOSDP.axiom_type, name="PrintfOwl_axiom_type", curie=DOSDP.curie('axiom_type'),
                   model_uri=DOSDP.PrintfOwl_axiom_type, domain=PrintfOwl, range=Optional[Union[str, "AxiomTypeOptions"]])

slots.PrintfOwl_text = Slot(uri=DOSDP.text, name="PrintfOwl_text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.PrintfOwl_text, domain=PrintfOwl, range=Optional[str])

slots.PrintfOwl_vars = Slot(uri=DOSDP.vars, name="PrintfOwl_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.PrintfOwl_vars, domain=PrintfOwl, range=Optional[Union[str, List[str]]])

slots.PrintfOwl_multi_clause = Slot(uri=DOSDP.multi_clause, name="PrintfOwl_multi_clause", curie=DOSDP.curie('multi_clause'),
                   model_uri=DOSDP.PrintfOwl_multi_clause, domain=PrintfOwl, range=Optional[str])

slots.PrintfOwlConvenience_annotations = Slot(uri=DOSDP.annotations, name="PrintfOwlConvenience_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.PrintfOwlConvenience_annotations, domain=PrintfOwlConvenience, range=Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]])

slots.PrintfOwlConvenience_text = Slot(uri=DOSDP.text, name="PrintfOwlConvenience_text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.PrintfOwlConvenience_text, domain=PrintfOwlConvenience, range=Optional[str])

slots.PrintfOwlConvenience_vars = Slot(uri=DOSDP.vars, name="PrintfOwlConvenience_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.PrintfOwlConvenience_vars, domain=PrintfOwlConvenience, range=Optional[Union[str, List[str]]])

slots.PrintfOwlConvenience_multi_clause = Slot(uri=DOSDP.multi_clause, name="PrintfOwlConvenience_multi_clause", curie=DOSDP.curie('multi_clause'),
                   model_uri=DOSDP.PrintfOwlConvenience_multi_clause, domain=PrintfOwlConvenience, range=Optional[str])

slots.RegexSub__in = Slot(uri=DOSDP._in, name="RegexSub__in", curie=DOSDP.curie('_in'),
                   model_uri=DOSDP.RegexSub__in, domain=RegexSub, range=Optional[str])

slots.RegexSub_out = Slot(uri=DOSDP.out, name="RegexSub_out", curie=DOSDP.curie('out'),
                   model_uri=DOSDP.RegexSub_out, domain=RegexSub, range=str)

slots.RegexSub_match = Slot(uri=DOSDP.match, name="RegexSub_match", curie=DOSDP.curie('match'),
                   model_uri=DOSDP.RegexSub_match, domain=RegexSub, range=str)

slots.RegexSub_sub = Slot(uri=DOSDP.sub, name="RegexSub_sub", curie=DOSDP.curie('sub'),
                   model_uri=DOSDP.RegexSub_sub, domain=RegexSub, range=str)

slots.OPA_edge = Slot(uri=DOSDP.edge, name="OPA_edge", curie=DOSDP.curie('edge'),
                   model_uri=DOSDP.OPA_edge, domain=OPA, range=Union[str, List[str]])

slots.OPA_annotations = Slot(uri=DOSDP.annotations, name="OPA_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.OPA_annotations, domain=OPA, range=Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]])

slots.OPA__not = Slot(uri=DOSDP._not, name="OPA__not", curie=DOSDP.curie('_not'),
                   model_uri=DOSDP.OPA__not, domain=OPA, range=Optional[Union[bool, Bool]])

slots.PrintfAnnotationObo_annotations = Slot(uri=DOSDP.annotations, name="PrintfAnnotationObo_annotations", curie=DOSDP.curie('annotations'),
                   model_uri=DOSDP.PrintfAnnotationObo_annotations, domain=PrintfAnnotationObo, range=Optional[Union[Union[dict, Annotation], List[Union[dict, Annotation]]]])

slots.PrintfAnnotationObo_xrefs = Slot(uri=DOSDP.xrefs, name="PrintfAnnotationObo_xrefs", curie=DOSDP.curie('xrefs'),
                   model_uri=DOSDP.PrintfAnnotationObo_xrefs, domain=PrintfAnnotationObo, range=Optional[str])

slots.PrintfAnnotationObo_text = Slot(uri=DOSDP.text, name="PrintfAnnotationObo_text", curie=DOSDP.curie('text'),
                   model_uri=DOSDP.PrintfAnnotationObo_text, domain=PrintfAnnotationObo, range=Optional[str])

slots.PrintfAnnotationObo_vars = Slot(uri=DOSDP.vars, name="PrintfAnnotationObo_vars", curie=DOSDP.curie('vars'),
                   model_uri=DOSDP.PrintfAnnotationObo_vars, domain=PrintfAnnotationObo, range=Optional[Union[str, List[str]]])

slots.PrintfAnnotationObo_multi_clause = Slot(uri=DOSDP.multi_clause, name="PrintfAnnotationObo_multi_clause", curie=DOSDP.curie('multi_clause'),
                   model_uri=DOSDP.PrintfAnnotationObo_multi_clause, domain=PrintfAnnotationObo, range=Optional[str])

slots.ListAnnotationObo_value = Slot(uri=DOSDP.value, name="ListAnnotationObo_value", curie=DOSDP.curie('value'),
                   model_uri=DOSDP.ListAnnotationObo_value, domain=ListAnnotationObo, range=str)

slots.ListAnnotationObo_xrefs = Slot(uri=DOSDP.xrefs, name="ListAnnotationObo_xrefs", curie=DOSDP.curie('xrefs'),
                   model_uri=DOSDP.ListAnnotationObo_xrefs, domain=ListAnnotationObo, range=Optional[str])



import click
from typing import Union, Dict, List, Any
from collections import defaultdict
import json
import yaml
import gzip

from dataclasses import dataclass

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

from schema_automator.generalizers.generalizer import Generalizer
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from linkml_runtime.utils.formatutils import camelcase

from schema_automator.utils.schemautils import write_schema


@dataclass
class JsonDataGeneralizer(Generalizer):
    mappings: dict = None
    omit_null: bool = None

    def convert(self, input: Union[str, Dict], format: str = 'json',
                container_class_name='Container',
                **kwargs) -> SchemaDefinition:
        csv_engine = CsvDataGeneralizer(**kwargs)

        if isinstance(input, dict):
            obj = input
        else:
            if format.endswith('.gz'):
                format = format.replace('.gz', '')
                stream = gzip.open(input)
            else:
                stream = open(input)

            with stream:
                if format == 'json':
                    obj = json.load(stream)
                elif format == 'yaml':
                    obj = yaml.safe_load(stream)
                else:
                    raise Exception(f'bad format {format}')
        rows_by_table = defaultdict(list)
        self.rows_by_table = rows_by_table
        self._convert_obj(obj, table=container_class_name)
        schemas = []
        for cn, rows_dict in rows_by_table.items():
            schema = csv_engine.convert_dicts(rows_dict, cn, cn)
            schemas.append(schema)
        sv = SchemaView(schemas[0])
        for s in schemas[1:]:
            sv.merge_schema(s)
        schema = sv.schema
        schema.classes[container_class_name].tree_root = True
        return schema

    def _key_to_classname(self, k: str) -> str:
        return camelcase(k)

    def _convert_obj(self, obj, table='Container'):
        if isinstance(obj, dict):
            row = defaultdict(set)
            for k, v in obj.items():
                if v is None and self.omit_null:
                    continue
                row[k] = self._convert_obj(v, table=camelcase(k))
            self.rows_by_table[table].append(row)
            return f'$ref:{table}'
        elif isinstance(obj, list):
            new_list = [self._convert_obj(v, table=table) for v in obj]
            return new_list
        else:
            return obj




    def _as_name(self, v):
        v = str(v)
        for sep in ['#', '/']:
            if sep in v:
                return v.split(sep)[-1]
        return v

def parse_frontmatter_files(paths: List[str], text_slot='_text') -> Any:
    blocks = []
    for path in paths:
        with open(path) as stream:
            state = 0
            yamlstr = ""
            txt = ""
            for line in stream.readlines():
                if line.startswith('---'):
                    state += 1
                else:
                    if state == 1:
                        yamlstr += line
                    elif state == 2:
                        txt += line
                    elif state > 2:
                        raise Exception(f'Limited to one frontmatter block per file')
            obj = yaml.safe_load(yamlstr)
            obj[text_slot] = txt
            blocks.append(obj)
    return blocks



@click.command()
@click.argument('input')
@click.option('--container-class-name', help="name of root class")
@click.option('--format', '-f', default='json', help="json or yaml (or json.gz or yaml.gz) or frontmatter")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def json2model(input, format, omit_null, **kwargs):
    """ Infer a model from JSON instance data


    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    schema = ie.convert(input, dir=dir, format=format, **kwargs)
    write_schema(schema)

@click.command()
@click.argument('inputs', nargs=-1)
@click.option('--container-class-name', help="name of root class")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def frontmatter2model(inputs, format, omit_null, **kwargs):
    """ Infer a model from frontmatter files


    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    objs = parse_frontmatter_files(list(inputs))
    schema = ie.convert({'objects': objs}, dir=dir, format=format, **kwargs)
    write_schema(schema)

if __name__ == '__main__':
    json2model()





import click
import pandas as pd
from linkml_runtime.utils.schemaview import SchemaView, Annotation
from linkml_runtime.dumpers import yaml_dumper

import logging
import click_log

# todo add logger
logger = logging.getLogger(__name__)
click_log.basic_config(logger)


# overwrites by default?
@click.command()
@click_log.simple_verbosity_option(logger)
@click.option('--tsv_in', type=click.Path(exists=True), required=True)
@click.option('--tsv_encoding', default="utf_16", show_default=True)
@click.option('--model_in', type=click.Path(exists=True), required=True)
@click.option('--curated_yaml', type=click.Path(), default="curated.yaml", show_default=True)
@click.option('--selected_enum', required=True)
def curated_to_enums(tsv_in, model_in, selected_enum, tsv_encoding, curated_yaml):
    from_tsv = pd.read_csv(tsv_in, sep="\t", encoding=tsv_encoding)
    from_tsv.index = from_tsv['text']
    logger.info(from_tsv)
    # check if an index appears more than once
    if from_tsv.index.is_unique:
        ft_dict = from_tsv.to_dict(orient="index")
    else:
        logger.error("index is not unique")
        exit()

    from_model = SchemaView(model_in)
    mschema = from_model.schema

    menum = from_model.get_enum(selected_enum)
    me_pvs = menum.permissible_values
    mep_keys = list(me_pvs.keys())
    mep_keys.sort()

    ft_keys = [i for i in list(ft_dict.keys()) if i == str(i)]
    ft_keys.sort()

    comparables = list(set(mep_keys).intersection(set(ft_keys)))
    comparables.sort()

    for i in comparables:
        logger.info(i)
        model_says = me_pvs[i]
        tsv_says = ft_dict[i]
        if not pd.isna(tsv_says['curated_id']) and not pd.isna(tsv_says['curated_pref_lab']) and not pd.isna(
                tsv_says['curated_type']) and not pd.isna(tsv_says['curated_val']):
            model_says.meaning = tsv_says['curated_id']
            model_says.title = tsv_says['curated_pref_lab']
            # todo delete them, don't set to empty strings
            model_says.annotations["match_id"] = tsv_says["curated_id"]
            model_says.annotations["match_pref_lab"] = tsv_says["curated_pref_lab"]
            model_says.annotations["match_type"] = tsv_says["curated_type"]
            model_says.annotations["match_val"] = tsv_says["curated_val"]
            model_says.annotations["curation_notes"] = tsv_says["curation_notes"]
            model_says.annotations["cosine"] = None
            model_says.annotations["curated"] = True
            me_pvs[i] = model_says

    menum.permissible_values = me_pvs

    meanings_tally = []
    for i in mep_keys:
        i_m = menum.permissible_values[i].meaning
        meanings_tally.append(i_m)

    # todo for collapsing pvs to get one with each meaning, the texts could be aliases or something like that?
    #   don't see an applicable slot at https://linkml.io/linkml-model/docs/PermissibleValue/#class-permissible_value

    pvs_per_meaning = pd.Series(meanings_tally).value_counts(dropna=False)

    for i in mep_keys:
        i_a = menum.permissible_values[i].annotations
        i_m = menum.permissible_values[i].meaning
        if i_m in pvs_per_meaning:
            # todo inconsistent annotation structure
            i_a["pvs_per_meaning"] = Annotation(tag="pvs_per_meaning", value=pvs_per_meaning[i_m])
            menum.permissible_values[i].annotations = i_a

    mschema.enums[selected_enum] = menum
    yaml_dumper.dump(mschema, curated_yaml)


if __name__ == '__main__':
    curated_to_enums()



from abc import ABC

from linkml_runtime.linkml_model import SchemaDefinition, Prefix


class Generalizer(ABC):
    """
    Abstract Base Class for all Generalization Engines.

    Generalization Engines take example data in some format and generalizes to a new SchemaDefinition
    """

    def convert(self, file: str, **kwargs) -> SchemaDefinition:
        raise NotImplementedError

    def add_prefix(self, schema: SchemaDefinition, prefix: str, url: str):
        schema.prefixes[prefix] = Prefix(prefix, url)

    def add_default_prefixes(self, schema: SchemaDefinition):
        self.add_prefix(schema, 'linkml', 'https://w3id.org/linkml/')


import logging
from copy import copy
from dataclasses import dataclass, field
from typing import List, Dict, Any, Tuple, Union, Optional

import click
from linkml.generators.pythongen import PythonGenerator
from linkml.utils import datautils
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinitionName, SlotDefinitionName
from linkml_runtime.utils.yamlutils import YAMLRoot

ID_REF = str
PATH = List[str]


@dataclass
class InstanceView:
    """
    A wrapper over an instance and it's sub-instances (instance tree)
    """
    root: YAMLRoot = None
    schemaview: SchemaView = None
    id_to_path: Dict[ID_REF, PATH] = field(default_factory=lambda: {})
    id_to_obj: Dict[ID_REF, YAMLRoot] = field(default_factory=lambda: {})
    references: List[Tuple[ID_REF, ID_REF]] = field(default_factory=lambda: [])

    def create_index(self, obj: Union[YAMLRoot, Dict, List] = None,
                     path=None, parent=None):
        """
        Creates internal index

        :param obj:
        :param path:
        :param parent:
        :return:
        """
        logging.debug(f'Creating index...')
        if path is None:
            path = []
        if obj is None:
            obj = self.root
        sv = self.schemaview
        indent = len(path) * '    '
        #print(f'{indent}{type(obj)} // {path}')
        if isinstance(obj, dict):
            for k, v in obj.items():
                self.create_index(v, path=path + [k], parent=parent)
        elif isinstance(obj, list):
            for i in range(0, len(obj)-1):
                self.create_index(obj[i], path=path + [i], parent=parent)
        elif isinstance(obj, YAMLRoot):
            cn = type(obj).class_name
            for slot in sv.class_induced_slots(cn):
                v = getattr(obj, slot.alias, None)
                if v is not None:
                    if slot.identifier:
                        self.references.append((parent, v))
                        # TODO: use explicit scheme to prioritize
                        if v not in self.id_to_path:
                            self.id_to_path[v] = path
                            self.id_to_obj[v] = obj
                        else:
                            curr_path = self.id_to_path[v]
                            if len(path) < len(curr_path):
                                self.id_to_path[v] = path
                                self.id_to_obj[v] = obj
                    else:
                        if slot.range in sv.all_classes():
                            if isinstance(v, str):
                                self.references.append((parent, v))
                        self.create_index(v, path=path+[slot.alias], parent=obj)
        logging.debug(f'Created index')

    def fetch_object(self, id_ref: ID_REF, default_val = None) -> Optional[YAMLRoot]:
        """
        Fetches an object by ID

        :param id_ref:
        :param default_val:
        :return: default_val if ID is dangling, otherwise return object with that ID
        """
        return self.id_to_obj.get(id_ref, default_val)


    def extract(self, seeds: List[ID_REF], preserve_slots: List[SlotDefinitionName] = None) -> YAMLRoot:
        """
        Extracts a subtree using a seed set of identifiers

        The subtree will include ancestors and descendants of identifiers, plus all recursive
        subtrees of all referenced entities

        :param seeds:
        :param preserve_slots:
        :return:
        """
        logging.info(f'Extracting from seed: {seeds}')
        id_to_path = self.id_to_path
        xdict = {}
        to_visit = copy(seeds)
        visited = set()
        while len(to_visit) > 0:
            nxt = to_visit.pop()
            if nxt in visited:
                continue
            logging.info(f'Extracting: {nxt}; {len(to_visit)} to go')
            if nxt not in id_to_path:
                logging.warning(f'No element with ID {nxt}')
                continue
            obj = self.fetch_object(nxt)
            path = self.id_to_path[nxt][0:-1]
            if len(path) != 1:
                continue
                #raise ValueError(f'{path} for {nxt}')
            #print(f'{nxt} = {rg} // {type(obj)}')
            islot = path[0]
            if islot not in xdict:
                if isinstance(getattr(self.root, islot), dict):
                    xdict[islot] = {}
                else:
                    xdict[islot] = []
            if isinstance(xdict[islot], dict):
                xdict[islot][nxt] = obj
            else:
                xdict[islot].append(obj)
            refs = self.get_object_references(obj)
            for ref in refs:
                if ref not in visited:
                    to_visit.append(ref)
            visited.add(nxt)
        if preserve_slots is None:
            preserve_slots = []
        root_cls = type(self.root)
        for slot in self.schemaview.class_induced_slots(root_cls.class_name):
            if slot.required:
                preserve_slots.append(slot.alias)
        for slot in preserve_slots:
            xdict[slot] = getattr(self.root, slot)
        #print(f'VISITED={visited}')
        #print(yaml_dumper.dumps(xdict))
        return root_cls(**xdict)


    def get_object_references(self, start_obj: Optional[Union[YAMLRoot, Dict, List]],
                              depth: int = 0) -> List[ID_REF]:
        logging.info(f'Fetching object references from {type(start_obj)}')
        if object is None:
            return []
        sv = self.schemaview
        refs = []
        indent = '    ' * depth
        #print(f'{indent}IN {type(obj)}')
        seeds = [start_obj]
        visited = set()
        while len(seeds) > 0:
            obj = seeds.pop()
            if isinstance(obj, YAMLRoot):
                for slot in sv.class_induced_slots(type(obj).class_name):
                    v = getattr(obj, slot.alias, None)
                    if v is not None:
                        if slot.identifier:
                            refs.append(v)
                        else:
                            rg = slot.range
                            if rg in sv.all_classes():
                                #print(f'{indent} - OBJ {type(obj)} {slot.alias} = {v} // {rg} // cumul={refs}')
                                if isinstance(v, list):
                                    seeds += v
                                elif isinstance(v, dict):
                                    seeds += v.values()
                                else:
                                    seeds.append(v)
            else:
                if obj not in visited:
                    refs.append(obj)
                    if obj in self.id_to_obj:
                        seeds.append(self.id_to_obj[obj])
                    visited.add(obj)
        #print(f'{indent}OUT {type(obj)} {cn} = {refs}')
        return list(set(refs))



@click.command()
@click.argument('elements', nargs=-1)
@click.option('--schema', '-s', help="Schema file")
@click.option('--input', '-i', help="Data input")
@click.option('--output', '-o', help="Path to saved yaml schema")
def cli(elements, input, schema, output, **args):
    """
    Extract from instance data
    """
    sv = SchemaView(schema)
    fmt = datautils._get_format(input)
    loader = datautils.get_loader(fmt)
    python_module = PythonGenerator(schema).compile_module()
    target_class = datautils.infer_root_class(sv)
    py_target_class = python_module.__dict__[target_class]
    root = loader.load(source=input,  target_class=py_target_class,)
    iv = InstanceView(root=root, schemaview=sv)
    iv.extract(elements)

if __name__ == '__main__':
    cli()






import json
import logging
from typing import Union, Dict, Optional

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition, Definition, Prefix, DefinitionName

JSON = Dict


def is_url(s: str) -> bool:
    if s is not None:
        return s.startswith('http')
    else:
        return False

class JsonLdAnnotator:
    """
    Annotates a schema using URIs derived from a JSON-LD file


    """

    def annotate(self, schema: Union[str, SchemaDefinition], jsonld_path: str):
        sv = SchemaView(schema)
        with open(jsonld_path) as f:
            jsonld = json.load(f)
        ctxt = jsonld['@context']
        for c in sv.all_classes().values():
            self.annotate_element(c, 'class_uri', ctxt)
        for s in sv.all_slots().values():
            self.annotate_element(s, 'slot_uri', ctxt)
        all_curies = []
        for c in sv.all_classes().values():
            if c.class_uri:
                all_curies.append(c.class_uri)
        for s in sv.all_slots().values():
            if s.slot_uri:
                all_curies.append(s.slot_uri)
        all_prefixes = set()
        for cu in all_curies:
            all_prefixes.add(cu.split(':')[0])
        print(f'ALL PREFIXES: {all_prefixes}')
        for k, v in ctxt.items():
            if k in all_prefixes:
                if k not in schema.prefixes:
                    url = None
                    if isinstance(v, str):
                        url = v
                    if is_url(url):
                        schema.prefixes[k] = Prefix(k, url)

    def annotate_element(self, el: Definition, mapping_slot: str, jsonld: JSON):
        if el.name in jsonld:
            curie = self.get_curie(el.name, jsonld)
            if curie:
                logging.info(f'ANNOTATING: {el.name} with {curie}')
                setattr(el, mapping_slot, curie)
            else:
                logging.error(f'Cannot annotate: {el.name}')


    def get_curie(self, name: DefinitionName, jsonld: JSON) -> Optional[str]:
        if name in jsonld:
            curie = jsonld[name]
            if isinstance(curie, str):
                if curie.startswith('@'):
                    return None
                else:
                    return curie
            elif isinstance(curie, dict):
                if '@id' in curie:
                    return curie['@id']





from copy import deepcopy

import click
import logging
import yaml
from typing import Union, Dict, Tuple, List, Any
from collections import defaultdict
import os
from csv import DictWriter

from dataclasses import dataclass

from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, SlotDefinition, EnumDefinition, TypeDefinition, AnonymousSlotExpression
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.utils.formatutils import camelcase, underscore

from schema_automator.importers.import_engine import ImportEngine
from schema_automator.dosdp.model import Pattern, Printf
from schema_automator.utils.schemautils import write_schema

ALIAS = str
CURIE = str

METACLASS = 'OntologyClass'
TEMPLATE_CLASS = 'OntologyClassTemplate'
GROUPING_CLASS = 'OntologyClassSubset'

@dataclass
class DOSDPImportEngine(ImportEngine):
    """
    An ImportEngine that imports Ontology Design Patterns specified as DOSDP Yaml into a LinkML schema

    See `DOSDPs <https://github.com/INCATools/dead_simple_owl_design_patterns>`_

    Every template maps to a LinkML class, the default name for a template Foo as FooTemplate

    """
    mappings: dict = None
    include_unmapped_annotations = False

    def load_dp(self, path) -> Pattern:
        with open(path) as stream:
            obj = yaml.safe_load(stream)
        if 'def' in obj:
            obj['definition'] = obj['def']
            del obj['def']
        return yaml_loader.load(obj, target_class=Pattern)

    def convert(self, files: str, range_as_enums = True, **kwargs) -> SchemaDefinition:
        """
        Converts one or more YAML files into a Schema

        :param files:
        :param range_as_enums: if True, then class ranges are mapped to Enums
        :param kwargs:
        :return:
        """
        patterns = [self.load_dp(file) for file in files]
        schema = SchemaDefinition(**kwargs)
        if not schema.default_prefix:
            schema.default_prefix = schema.name
            schema.prefixes[schema.name] = f'{schema.id}'
        schema.default_curi_maps = ['semweb_context', 'obo_context']
        self.schema = schema
        mc = ClassDefinition(METACLASS)
        schema.classes[mc.name] = mc
        mc.description = "Instance of OWL Class"
        mc.aliases = ['OWL Class']
        mc.class_uri = 'owl:Class'
        mc.mixin = True
        mc.slots = ['id']
        tmc = ClassDefinition(TEMPLATE_CLASS)
        schema.classes[tmc.name] = tmc
        tmc.description = "Instances of OWL classes that conform to a template"
        tmc.aliases = ['design pattern record', 'template row']
        tmc.mixin = True
        tmc.mixins = [mc.name]
        gmc = ClassDefinition(GROUPING_CLASS)
        schema.classes[gmc.name] = gmc
        gmc.description = "Mixin for instances of OWL Classes that are used as range references"
        gmc.mixin = True
        gmc.mixins = [mc.name]
        schema.types['string'] = TypeDefinition(name='string', base='str', uri='xsd:string')
        def add_slot(s: SlotDefinition):
            schema.slots[s.name] = s
        add_slot(SlotDefinition('id',
                                identifier=True,
                                #range='uriorcurie',
                                description='Unique identifier for template instance'))
        add_slot(SlotDefinition('name',
                                slot_uri='rdfs:label',
                                description='Human readable label for template instance'))
        add_slot(SlotDefinition('definition',
                                slot_uri='IAO:0000115',
                                description='Human readable textual definition for template instance'))
        add_slot(SlotDefinition('subclass_of',
                                multivalued=True,
                                slot_uri='rdfs:subclass_of',
                                range=METACLASS))
        add_slot(SlotDefinition('equivalentTo',
                                description='String serialization of RHS expression in equivalence axiom',
                                close_mappings=['owl:equivalentClass']))
        licenses = set()
        for p in patterns:
            c = self.create_class(p, range_as_enums=range_as_enums)
            schema.classes[c.name] = c
        return schema

    def create_class(self, pattern: Pattern, range_as_enums: bool = True,
                     denormalize_labels: bool = True) -> ClassDefinition:
        """
        Create a LinkML class from a template/pattern

        The class will have name FooTemplate, and will mixin TemplateClass,
        and will have one slot per var (slots will also be inherited from TemplateClass)

        by default, the range of each var will be a class RangeClass, unless treated as enums

        :param pattern: DOSDP template
        :param range_as_enums:  if true, ranges will be enums
        :param denoramlized:  if true, create an optional label for each var
        :return:
        """
        def deref_class(name: ALIAS) -> CURIE:
            return self._deref(name, pattern.classes)

        schema = self.schema
        template_name = f'{pattern.pattern_name} template'
        cls = ClassDefinition(name=camelcase(template_name),
                              aliases=[template_name],
                              #title=template_name,  ## TODO: fix owlgen
                              description=pattern.description,
                              mixins=[TEMPLATE_CLASS])
        for contrib in pattern.contributors:
            cls.comments.append(f'Contributor: {contrib}')  ## TODO: extend metamodel
        extra_classes = []
        for vn, v in pattern.vars.items():
            var_range_as_curie = v.range
            for alias, alias_obj in pattern.classes.items():
                if f"'{alias}'" == v.range:
                    var_range_as_curie = alias_obj.curie
            range_base_name = self._name(v.range)
            if range_as_enums:
                rangedef = self.create_enum(f'{range_base_name} enum')
            else:
                rangedef = ClassDefinition(camelcase(f'{range_base_name} class'))
                #rangedef.is_a = METACLASS
                rangedef.mixins = [GROUPING_CLASS]
                subc_slot = SlotDefinition('subclass_of',
                                           has_member=AnonymousSlotExpression(equals_string=var_range_as_curie))
                rangedef.slot_usage[subc_slot.name] = subc_slot
                extra_classes.append(rangedef)
            rangedef.description = f'Any subclass of {var_range_as_curie} ({v.range})'
            slot = SlotDefinition(vn, range=rangedef.name)
            schema.slots[slot.name] = slot
            cls.slots.append(slot.name)
            cls.slot_usage[slot.name] = slot
            if denormalize_labels:
                label_slot = SlotDefinition(f'{vn}_label',
                                            description=f'label for {vn}',
                                            string_serialization=f'{{{vn}}}.label')
                schema.slots[label_slot.name] = label_slot
                cls.slots.append(label_slot.name)
                cls.slot_usage[label_slot.name] = label_slot
        # populate metaclass
        mc = schema.classes[METACLASS]

        for k in ['name', 'definition', 'equivalentTo']:
            pf = getattr(pattern, k)
            #print(f'k={k} rule={srule}')
            if pf is not None:
                slot = self._serialization_slot(k, pf)
                if slot.name not in mc.slots:
                    mc.slots.append(slot.name)
                    #schema.slots[slot.name] = deepcopy(slot)
                    #schema.slots[slot.name].string_serialization = None
                cls.slot_usage[slot.name] = slot
        # append these at the end
        for c in extra_classes:
            schema.classes[c.name] = c
        return cls

    def create_enum(self, name: str):
        name = underscore(self._name(name))
        enumdef = EnumDefinition(name)
        self.schema.enums[enumdef.name] = enumdef
        return enumdef


    def _serialization_slot(self, k: str, pf: Printf):
        slot = SlotDefinition(k)
        t = pf.text
        for vn in pf.vars:
            t = t.replace('%', f'{{{vn}}}', 1)
        slot.string_serialization = t
        return slot

    def _name(self, name: ALIAS) -> ALIAS:
        if name.startswith("'"):
            name = name.replace("'","")
        return name

    def _deref(self, name: ALIAS, index: dict) -> CURIE:
        if name.startswith("'"):
            name = name.replace("'","")
            return index[name].curie


@click.command()
@click.argument('dpfiles', nargs=-1) ## input DOSDPs
@click.option('--name', '-n', help="Schema name")
@click.option('--range-as-enum/--no-range-as-enums', default=True, help="Model range ontology classes as enyms")
@click.option('--output', '-o', help="Path to saved yaml schema")
def dosdp2model(dpfiles, output, **args):
    """
    Infer a model from DOSDP

    """
    ie = DOSDPImportEngine()
    schema = ie.convert(dpfiles, **args)
    write_schema(schema, output)

if __name__ == '__main__':
    dosdp2model()





# Auto generated from jsonschema.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-08-27 16:13
# Schema: jsonschema
#
# id: https://w3id.org/linkml/jsonschema
# description: A jsonschemamodel for defining linked open data schemas
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import String

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
OIO = CurieNamespace('OIO', 'http://www.geneontology.org/formats/oboInOwl#')
BIBO = CurieNamespace('bibo', 'http://purl.org/ontology/bibo/')
DCTERMS = CurieNamespace('dcterms', 'http://purl.org/dc/terms/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OSLC = CurieNamespace('oslc', 'http://open-services.net/ns/core#')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
PAV = CurieNamespace('pav', 'http://purl.org/pav/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = LINKML


# Types
class Ref(str):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "Ref"
    type_model_uri = LINKML.Ref


# Class references
class DefinitionName(extended_str):
    pass


@dataclass
class Schema(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LINKML.Schema
    class_class_curie: ClassVar[str] = "linkml:Schema"
    class_name: ClassVar[str] = "Schema"
    class_model_uri: ClassVar[URIRef] = LINKML.Schema

    definitions: Optional[Union[Dict[Union[str, DefinitionName], Union[dict, "Definition"]], List[Union[dict, "Definition"]]]] = empty_dict()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        self._normalize_inlined_as_dict(slot_name="definitions", slot_type=Definition, key_name="name", keyed=True)

        super().__post_init__(**kwargs)


@dataclass
class Definition(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LINKML.Definition
    class_class_curie: ClassVar[str] = "linkml:Definition"
    class_name: ClassVar[str] = "Definition"
    class_model_uri: ClassVar[URIRef] = LINKML.Definition

    name: Union[str, DefinitionName] = None
    title: Optional[str] = None
    type: Optional[str] = None
    properties: Optional[Union[Dict[Union[str, DefinitionName], Union[dict, "Definition"]], List[Union[dict, "Definition"]]]] = empty_dict()
    items: Optional[Union[Dict[Union[str, DefinitionName], Union[dict, "Definition"]], List[Union[dict, "Definition"]]]] = empty_dict()
    references: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, DefinitionName):
            self.name = DefinitionName(self.name)

        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        if self.type is not None and not isinstance(self.type, str):
            self.type = str(self.type)

        self._normalize_inlined_as_dict(slot_name="properties", slot_type=Definition, key_name="name", keyed=True)

        self._normalize_inlined_as_dict(slot_name="items", slot_type=Definition, key_name="name", keyed=True)

        if self.references is not None and not isinstance(self.references, str):
            self.references = str(self.references)

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.name = Slot(uri=LINKML.name, name="name", curie=LINKML.curie('name'),
                   model_uri=LINKML.name, domain=None, range=URIRef)

slots.definitions = Slot(uri=LINKML.definitions, name="definitions", curie=LINKML.curie('definitions'),
                   model_uri=LINKML.definitions, domain=None, range=Optional[Union[Dict[Union[str, DefinitionName], Union[dict, Definition]], List[Union[dict, Definition]]]])

slots.references = Slot(uri=LINKML.references, name="references", curie=LINKML.curie('references'),
                   model_uri=LINKML.references, domain=None, range=Optional[str])

slots.title = Slot(uri=LINKML.title, name="title", curie=LINKML.curie('title'),
                   model_uri=LINKML.title, domain=None, range=Optional[str])

slots.type = Slot(uri=LINKML.type, name="type", curie=LINKML.curie('type'),
                   model_uri=LINKML.type, domain=None, range=Optional[str])

slots.properties = Slot(uri=LINKML.properties, name="properties", curie=LINKML.curie('properties'),
                   model_uri=LINKML.properties, domain=None, range=Optional[Union[Dict[Union[str, DefinitionName], Union[dict, Definition]], List[Union[dict, Definition]]]])

slots.items = Slot(uri=LINKML.items, name="items", curie=LINKML.curie('items'),
                   model_uri=LINKML.items, domain=None, range=Optional[Union[Dict[Union[str, DefinitionName], Union[dict, Definition]], List[Union[dict, Definition]]]])



import click
from collections import defaultdict
import os
from csv import DictWriter
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

from rdflib import Graph, URIRef
from rdflib.namespace import RDF

from dataclasses import dataclass

from schema_automator.generalizers.generalizer import Generalizer
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from schema_automator.utils.schemautils import write_schema


@dataclass
class RdfDataGeneralizer(Generalizer):
    mappings: dict = None

    def convert(self, file: str, dir: str, **kwargs) -> SchemaDefinition:
        csv_engine = CsvDataGeneralizer()

        g = Graph()
        g.parse(file, **kwargs)
        self.mappings = {}
        paths = self.graph_to_tables(g, dir)
        schemas = []
        for c, tsvfile in paths.items():
            schemas.append(csv_engine.convert(tsvfile, class_name=c))
        sv = SchemaView(schemas[0])
        for s in schemas[1:]:
            sv.merge_schema(s)
        schema = sv.schema
        mappings = self.mappings
        for cn, c in schema.classes.items():
            if cn in mappings:
                c.class_uri = mappings[cn]
        for sn, s in schema.slots.items():
            if sn in mappings:
                s.slot_uri = mappings[sn]
        for en, e in schema.enums.items():
            for pvn, pvo in e.permissible_values.items():
                if pvn in mappings:
                    pvo.meaning = mappings[pvn]
        return schema

    def graph_to_tables(self, g: Graph, dir: str):
        mappings = self.mappings
        rows_by_table = defaultdict(list)
        for s,_,t_uriref in g.triples((None, RDF.type, None)):
            t = self._as_name(t_uriref)
            mappings[t] = str(t_uriref)
            #print(f'I={s} type={t}')
            row = defaultdict(set)
            row['id'] = {s}
            for _,p_uriref,o_uriref in g.triples((s, None, None)):
                p = self._as_name(p_uriref)
                o = self._as_name(o_uriref)
                mappings[p] = str(p_uriref)
                if isinstance(o_uriref, URIRef):
                    mappings[o] = str(o_uriref)
                row[p].add(o)
            rows_by_table[t].append(row)
            #print(f'  ROW={row}')
        paths = {}
        for t, rows in rows_by_table.items():
            path = f'{os.path.join(dir, t)}.tsv'
            paths[t] = path
            fields = []
            for row in rows:
                for k in row.keys():
                    if k not in fields:
                        fields.append(k)
                    v = list(row[k])
                    v = '|'.join(v)
                    row[k] = v
            print(f'Writing {len(rows)} to {path}')
            with open(path, 'w') as stream:
                w = DictWriter(stream, delimiter='\t', fieldnames=fields)
                w.writeheader()
                for row in rows:
                    w.writerow(row)
        return paths

    def _as_name(self, v):
        v = str(v)
        for sep in ['#', '/']:
            if sep in v:
                return v.split(sep)[-1]
        return v

@click.command()
@click.argument('rdffile')
@click.option('--dir', '-d', required=True)
def rdf2model(rdffile, dir, **args):
    """ Infer a model from RDF instance data """
    sie = RdfDataGeneralizer()
    if not os.path.exists(dir):
        os.makedirs(dir)
    schema = sie.convert(rdffile, dir=dir, format='ttl')
    write_schema(schema)

if __name__ == '__main__':
    rdf2model()


